# -*- coding: utf-8 -*-
#
# This file is part of CodraFT Project
# https://codra-ingenierie-informatique.github.io/CodraFT/
#
# Licensed under the terms of the BSD 3-Clause or the CeCILL-B License
# (see codraft/LICENSE for details)

"""
JSON files (.json)
"""

# pylint: disable=invalid-name  # Allows short reference names like x, y, ...

from __future__ import annotations

import json
import os
from collections.abc import Callable, Sequence
from typing import Any
from uuid import uuid1

import numpy as np

from guidata.io.base import BaseIOHandler, WriterMixin


class CustomJSONEncoder(json.JSONEncoder):
    """Custom JSON Encoder"""

    def default(self, o: Any) -> Any:
        """Override JSONEncoder method"""
        if isinstance(o, np.ndarray):
            olist = o.tolist()
            if o.dtype in (np.complex64, np.complex128):
                olist = o.real.tolist() + o.imag.tolist()
            return ["array", olist, str(o.dtype)]
        if isinstance(o, np.generic):
            if isinstance(o, np.integer):
                return int(o)
            try:
                return float(o)
            except ValueError:
                return str(o)
        if isinstance(o, bytes):
            return o.decode()
        return json.JSONEncoder.default(self, o)


class CustomJSONDecoder(json.JSONDecoder):
    """Custom JSON Decoder"""

    def __init__(self, *args, **kwargs) -> None:
        json.JSONDecoder.__init__(self, object_hook=self.object_hook, *args, **kwargs)

    def __iterate_dict(self, obj: Any) -> Any:
        """Iterate dictionaries"""
        if isinstance(obj, list):
            if len(obj) == 3:
                family, data, dtypestr = obj
                try:
                    dtype = np.dtype(dtypestr)
                    if family == "array":
                        if dtype in (np.complex64, np.complex128):
                            return np.asarray(
                                data[: len(data) // 2], dtype
                            ) + 1j * np.asarray(data[len(data) // 2 :], dtype)
                        return np.asarray(data, dtype)
                except (TypeError, ValueError):
                    pass
            return [self.__iterate_dict(item) for item in obj]
        elif isinstance(obj, dict):
            for key, value in list(obj.items()):
                obj[key] = self.__iterate_dict(value)
        return obj

    def object_hook(self, obj: dict) -> dict:  # pylint: disable=E0202
        """Object hook"""
        for key, value in list(obj.items()):
            obj[key] = self.__iterate_dict(value)
        return obj


class JSONHandler(BaseIOHandler):
    """Class handling JSON r/w

    Args:
        filename: JSON filename (if None, use `jsontext` attribute)
    """

    def __init__(self, filename: str | None = None) -> None:
        super().__init__()
        self.jsondata = {}
        self.jsontext: str | None = None
        self.filename = filename

    def get_parent_group(self) -> dict:
        """Get parent group"""
        parent = self.jsondata
        for option in self.option[:-1]:
            parent = parent.setdefault(option, {})
        return parent

    def set_json_dict(self, jsondata: dict) -> None:
        """Set JSON data dictionary

        Args:
            jsondata: JSON data dictionary
        """
        self.jsondata = jsondata

    def get_json_dict(self) -> dict:
        """Return JSON data dictionary"""
        return self.jsondata

    def get_json(self, indent: int | None = None) -> str | None:
        """Get JSON string

        Args:
            indent: Indentation level

        Returns:
            JSON string
        """
        if self.jsondata is not None:
            return json.dumps(self.jsondata, indent=indent, cls=CustomJSONEncoder)
        return None

    def load(self) -> None:
        """Load JSON file"""
        if self.filename is not None:
            with open(self.filename, mode="rb") as fdesc:
                self.jsontext = fdesc.read().decode()
        self.jsondata = json.loads(self.jsontext, cls=CustomJSONDecoder)

    def save(self, path: str | None = None) -> None:
        """Save JSON file

        Args:
            path: Path to save the JSON file (if None, implies current directory)
        """
        if self.filename is not None:
            filepath = self.filename
            if path:
                filepath = os.path.join(path, filepath)
            with open(filepath, mode="wb") as fdesc:
                fdesc.write(self.get_json(indent=4).encode())

    def close(self) -> None:
        """Expected close method: do nothing for JSON I/O handler classes"""


class JSONWriter(JSONHandler, WriterMixin):
    """Class handling JSON serialization"""

    def write_any(self, val) -> None:
        """Write any value type"""
        group = self.get_parent_group()
        group[self.option[-1]] = val

    def write_none(self) -> None:
        """Write None"""
        self.write_any(None)

    write_sequence = write_dict = write_str = write_bool = write_int = write_float = (
        write_array
    ) = write_any

    def write_object_list(self, seq: Sequence[Any] | None, group_name: str) -> None:
        """
        Write an object sequence to the HDF5 file in a group.
        Objects must implement the DataSet-like `serialize` method.

        Args:
            seq: The object sequence to write. Defaults to None.
            group_name: The name of the group in which to write the objects.
        """
        with self.group(group_name):
            if seq is None:
                self.write_none()
            else:
                ids = []
                for obj in seq:
                    guid = str(uuid1())
                    ids.append(guid)
                    with self.group(guid):
                        if obj is None:
                            self.write_none()
                        else:
                            obj.serialize(self)
                self.write(ids, "IDs")


class NoDefault:
    """Class to represent the absence of a default value."""

    pass


class JSONReader(JSONHandler):
    """Class handling JSON deserialization

    Args:
        fname_or_jsontext: JSON filename or JSON text
    """

    def __init__(self, fname_or_jsontext: str) -> None:
        """JSONReader constructor"""
        JSONHandler.__init__(self, fname_or_jsontext)
        if fname_or_jsontext is not None and not os.path.isfile(fname_or_jsontext):
            self.filename = None
            self.jsontext = fname_or_jsontext
        self.load()

    def read(
        self,
        group_name: str | None = None,
        func: Callable[[], Any] | None = None,
        instance: Any | None = None,
        default: Any | NoDefault = NoDefault,
    ) -> Any:
        """
        Read a value from the current group or specified group_name.

        Args:
            group_name: The name of the group to read from. Defaults to None.
            func: The function to use for reading the value. Defaults to None.
            instance: An object that implements the DataSet-like `deserialize` method.
             Defaults to None.
            default: The default value to return if the value is not found.
             Defaults to `NoDefault` (no default value: raises an exception if the
             value is not found).

        Returns:
            The read value.
        """
        if group_name:
            self.begin(group_name)
        try:
            if instance is None:
                if func is None:
                    func = self.read_any
                val = func()
            else:
                group = self.get_parent_group()
                if group_name not in group:
                    # This is an attribute (not a group), meaning that
                    # the object was None when deserializing it
                    val = None
                else:
                    instance.deserialize(self)
                    val = instance
        except Exception:  # pylint:disable=broad-except
            if default is NoDefault:
                raise
            val = default
        if group_name:
            self.end(group_name)
        return val

    def read_any(self) -> Any:
        """Read any value type"""
        group = self.get_parent_group()
        return group[self.option[-1]]

    def read_object_list(
        self,
        group_name: str,
        klass: type[Any],
        progress_callback: Callable[[int], bool] | None = None,
    ) -> list[Any]:
        """Read an object sequence from a group.

        Objects must implement the DataSet-like `deserialize` method.
        `klass` is the object class which constructor requires no argument.

        Args:
            group_name: The name of the group to read the object sequence from.
            klass: The object class which constructor requires no argument.
            progress_callback: A function to call with an integer argument (progress:
             0 --> 100). The function returns the `cancel` state (True: progress
             dialog has been canceled, False otherwise).
        """
        with self.group(group_name):
            try:
                ids = self.read("IDs", func=self.read_sequence)
            except ValueError:
                # None was saved instead of list of objects
                self.end("IDs")
                return None
            seq = []
            count = len(ids)
            for idx, name in enumerate(ids):
                if progress_callback is not None:
                    if progress_callback(int(100 * float(idx) / count)):
                        break
                with self.group(name):
                    group = self.get_parent_group()
                    if group[name] is None:
                        # The object was None when deserializing it
                        obj = None
                    else:
                        obj = klass()
                        obj.deserialize(self)
                seq.append(obj)
        return seq

    read_unicode = read_sequence = read_dict = read_float = read_int = read_str = (
        read_bool
    ) = read_array = read_any



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause
# (see guidata/LICENSE for details)
#
# The array editor subpackage was derived from Spyder's arrayeditor.py module
# which is licensed under the terms of the MIT License (see spyder/__init__.py
# for details), copyright © Spyder Project Contributors

"""guidata.widgets.arrayeditor
===========================

This package provides a NumPy Array Editor Dialog based on Qt.

.. autoclass:: ArrayEditor
    :show-inheritance:
    :members:

"""

from .arrayeditor import ArrayEditor  # noqa: F401



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""Shell widgets: base, python and terminal"""

# pylint: disable=C0103
# pylint: disable=R0903
# pylint: disable=R0911
# pylint: disable=R0201

import builtins
import keyword
import locale
import os
import os.path as osp
import re
import sys
import time

from qtpy.compat import getsavefilename
from qtpy.QtCore import Property, QCoreApplication, Qt, QTimer, Signal, Slot
from qtpy.QtGui import QKeyEvent, QTextCharFormat, QTextCursor
from qtpy.QtWidgets import QApplication, QMenu, QToolTip

from guidata.config import CONF, _
from guidata.configtools import get_icon
from guidata.qthelpers import add_actions, create_action, keybinding
from guidata.utils import encoding
from guidata.widgets.console.base import ConsoleBaseWidget
from guidata.widgets.console.mixins import (
    BrowseHistoryMixin,
    GetHelpMixin,
    SaveHistoryMixin,
    TracebackLinksMixin,
)

DEBUG = False
STDERR = sys.stderr


def tuple2keyevent(past_event):
    """Convert tuple into a QKeyEvent instance"""
    return QKeyEvent(*past_event)


def restore_keyevent(event):
    """

    :param event:
    :return:
    """
    if isinstance(event, tuple):
        _, key, modifiers, text, _, _ = event
        event = tuple2keyevent(event)
    else:
        text = event.text()
        modifiers = event.modifiers()
        key = event.key()
    ctrl = modifiers & Qt.ControlModifier
    shift = modifiers & Qt.ShiftModifier
    return event, text, key, ctrl, shift


class ShellBaseWidget(ConsoleBaseWidget, SaveHistoryMixin, BrowseHistoryMixin):
    """
    Shell base widget
    """

    redirect_stdio = Signal(bool)
    sig_keyboard_interrupt = Signal()
    execute = Signal(str)
    append_to_history = Signal(str, str)

    def __init__(self, parent, profile=False, initial_message=None, read_only=False):
        """
        `parent` : specifies the parent widget
        """
        ConsoleBaseWidget.__init__(self, parent)
        SaveHistoryMixin.__init__(self)
        BrowseHistoryMixin.__init__(self)

        # Read-only console is not the more used case, but it may be useful in
        # some cases (e.g. when using the console as a log window)
        self.setReadOnly(read_only)

        self.historylog_filename = "history.log"

        # Prompt position: tuple (line, index)
        self.current_prompt_pos = None
        self.new_input_line = True

        # Context menu
        self.menu = None
        self.setup_context_menu()

        # Simple profiling test
        self.profile = profile

        # Buffer to increase performance of write/flush operations
        self.__buffer = []
        if initial_message:
            self.__buffer.append(initial_message)

        self.__timestamp = 0.0
        self.__flushtimer = QTimer(self)
        self.__flushtimer.setSingleShot(True)
        self.__flushtimer.timeout.connect(self.flush)

        # Give focus to widget
        self.setFocus()

        # Cursor width
        self.setCursorWidth(CONF.get("console", "cursor/width"))

    def toggle_wrap_mode(self, enable):
        """Enable/disable wrap mode"""
        self.set_wrap_mode("character" if enable else None)

    def set_font(self, font):
        """Set shell styles font"""
        self.setFont(font)
        self.set_pythonshell_font(font)
        cursor = self.textCursor()
        cursor.select(QTextCursor.Document)
        charformat = QTextCharFormat()
        charformat.setFontFamilies([font.family()])
        charformat.setFontPointSize(font.pointSize())
        cursor.mergeCharFormat(charformat)

    # ------ Context menu
    def setup_context_menu(self):
        """Setup shell context menu"""
        self.menu = QMenu(self)
        self.cut_action = create_action(
            self,
            _("Cut"),
            shortcut=keybinding("Cut"),
            icon=get_icon("editcut.png"),
            triggered=self.cut,
        )
        self.copy_action = create_action(
            self,
            _("Copy"),
            shortcut=keybinding("Copy"),
            icon=get_icon("editcopy.png"),
            triggered=self.copy,
        )
        paste_action = create_action(
            self,
            _("Paste"),
            shortcut=keybinding("Paste"),
            icon=get_icon("editpaste.png"),
            triggered=self.paste,
        )
        save_action = create_action(
            self,
            _("Save history log..."),
            icon=get_icon("filesave.png"),
            tip=_(
                "Save current history log (i.e. all inputs and outputs) in a text file"
            ),
            triggered=self.save_historylog,
        )
        self.delete_action = create_action(
            self,
            _("Delete"),
            shortcut=keybinding("Delete"),
            icon=get_icon("editdelete.png"),
            triggered=self.delete,
        )
        selectall_action = create_action(
            self,
            _("Select All"),
            shortcut=keybinding("SelectAll"),
            icon=get_icon("selectall.png"),
            triggered=self.selectAll,
        )
        add_actions(
            self.menu,
            (
                self.cut_action,
                self.copy_action,
                paste_action,
                self.delete_action,
                None,
                selectall_action,
                None,
                save_action,
            ),
        )

    def contextMenuEvent(self, event):
        """Reimplement Qt method"""
        state = self.has_selected_text()
        self.copy_action.setEnabled(state)
        self.cut_action.setEnabled(state)
        self.delete_action.setEnabled(state)
        self.menu.popup(event.globalPos())
        event.accept()

    # ------ Input buffer
    def get_current_line_from_cursor(self):
        """

        :return:
        """
        return self.get_text("cursor", "eof")

    def _select_input(self):
        """Select current line (without selecting console prompt)"""
        line, index = self.get_position("eof")
        if self.current_prompt_pos is None:
            pline, pindex = line, index
        else:
            pline, pindex = self.current_prompt_pos
        self.setSelection(pline, pindex, line, index)

    @Slot()
    def clear_terminal(self):
        """
        Clear terminal window
        Child classes reimplement this method to write prompt
        """
        self.clear()

    # The buffer being edited
    def _set_input_buffer(self, text):
        """Set input buffer"""
        if self.current_prompt_pos is not None:
            self.replace_text(self.current_prompt_pos, "eol", text)
        else:
            self.insert(text)
        self.set_cursor_position("eof")

    def _get_input_buffer(self):
        """Return input buffer"""
        input_buffer = ""
        if self.current_prompt_pos is not None:
            input_buffer = self.get_text(self.current_prompt_pos, "eol")
            input_buffer = input_buffer.replace(os.linesep, "\n")
        return input_buffer

    input_buffer = Property("QString", _get_input_buffer, _set_input_buffer)

    # ------ Prompt
    def new_prompt(self, prompt):
        """
        Print a new prompt and save its (line, index) position
        """
        if self.get_cursor_line_column()[1] != 0:
            self.write("\n")
        self.write(prompt, prompt=True)
        # now we update our cursor giving end of prompt
        self.current_prompt_pos = self.get_position("cursor")
        self.ensureCursorVisible()
        self.new_input_line = False

    def check_selection(self):
        """
        Check if selected text is r/w,
        otherwise remove read-only parts of selection
        """
        if self.current_prompt_pos is None:
            self.set_cursor_position("eof")
        else:
            self.truncate_selection(self.current_prompt_pos)

    # ------ Copy / Keyboard interrupt
    @Slot()
    def copy(self):
        """Copy text to clipboard... or keyboard interrupt"""
        if self.has_selected_text():
            ConsoleBaseWidget.copy(self)
        elif not sys.platform == "darwin":
            self.interrupt()

    def interrupt(self):
        """Keyboard interrupt"""
        self.sig_keyboard_interrupt.emit()

    @Slot()
    def cut(self):
        """Cut text"""
        self.check_selection()
        if self.has_selected_text():
            ConsoleBaseWidget.cut(self)

    @Slot()
    def delete(self):
        """Remove selected text"""
        self.check_selection()
        if self.has_selected_text():
            ConsoleBaseWidget.remove_selected_text(self)

    @Slot()
    def save_historylog(self):
        """Save current history log (all text in console)"""
        title = _("Save history log")
        self.redirect_stdio.emit(False)
        filename, _selfilter = getsavefilename(
            self, title, self.historylog_filename, "%s (*.log)" % _("History logs")
        )
        self.redirect_stdio.emit(True)
        if filename:
            filename = osp.normpath(filename)
            try:
                encoding.write(str(self.get_text_with_eol()), filename)
                self.historylog_filename = filename
            except EnvironmentError:
                pass

    # ------ Basic keypress event handler
    def on_enter(self, command):
        """on_enter"""
        self.execute_command(command)

    def execute_command(self, command):
        """

        :param command:
        """
        self.execute.emit(command)
        self.add_to_history(command)
        self.new_input_line = True

    def on_new_line(self):
        """On new input line"""
        self.set_cursor_position("eof")
        self.current_prompt_pos = self.get_position("cursor")
        self.new_input_line = False

    @Slot()
    def paste(self):
        """Reimplemented slot to handle multiline paste action"""
        if self.new_input_line:
            self.on_new_line()
        ConsoleBaseWidget.paste(self)

    def keyPressEvent(self, event):
        """
        Reimplement Qt Method
        Basic keypress event handler
        (reimplemented in InternalShell to add more sophisticated features)
        """
        if self.preprocess_keyevent(event):
            # Event was accepted in self.preprocess_keyevent
            return
        self.postprocess_keyevent(event)

    def preprocess_keyevent(self, event):
        """Pre-process keypress event:
        return True if event is accepted, false otherwise"""
        # Copy must be done first to be able to copy read-only text parts
        # (otherwise, right below, we would remove selection
        #  if not on current line)
        ctrl = event.modifiers() & Qt.ControlModifier
        meta = event.modifiers() & Qt.MetaModifier  # meta=ctrl in OSX
        if event.key() == Qt.Key_C and (
            (Qt.MetaModifier | Qt.ControlModifier) & event.modifiers()
        ):
            if meta and sys.platform == "darwin":
                self.interrupt()
            elif ctrl:
                self.copy()
            event.accept()
            return True

        if self.new_input_line and (
            len(event.text())
            or event.key() in (Qt.Key_Up, Qt.Key_Down, Qt.Key_Left, Qt.Key_Right)
        ):
            self.on_new_line()

        return False

    def postprocess_keyevent(self, event):
        """Post-process keypress event:
        in InternalShell, this is method is called when shell is ready"""
        event, text, key, ctrl, shift = restore_keyevent(event)

        # Is cursor on the last line? and after prompt?
        if len(text):
            # XXX: Shouldn't it be: `if len(unicode(text).strip(os.linesep))` ?
            if self.has_selected_text():
                self.check_selection()
            self.restrict_cursor_position(self.current_prompt_pos, "eof")

        cursor_position = self.get_position("cursor")

        if key in (Qt.Key_Return, Qt.Key_Enter):
            if self.is_cursor_on_last_line():
                self._key_enter()
            # add and run selection
            else:
                self.insert_text(self.get_selected_text(), at_end=True)

        elif key == Qt.Key_Insert and not shift and not ctrl:
            self.setOverwriteMode(not self.overwriteMode())

        elif key == Qt.Key_Delete:
            if self.has_selected_text():
                self.check_selection()
                self.remove_selected_text()
            elif self.is_cursor_on_last_line():
                self.stdkey_clear()

        elif key == Qt.Key_Backspace:
            self._key_backspace(cursor_position)

        elif key == Qt.Key_Tab:
            self._key_tab()

        elif key == Qt.Key_Space and ctrl:
            self._key_ctrl_space()

        elif key == Qt.Key_Left:
            if self.current_prompt_pos == cursor_position:
                # Avoid moving cursor on prompt
                return
            method = (
                self.extend_selection_to_next if shift else self.move_cursor_to_next
            )
            method("word" if ctrl else "character", direction="left")

        elif key == Qt.Key_Right:
            if self.is_cursor_at_end():
                return
            method = (
                self.extend_selection_to_next if shift else self.move_cursor_to_next
            )
            method("word" if ctrl else "character", direction="right")

        elif (key == Qt.Key_Home) or ((key == Qt.Key_Up) and ctrl):
            self._key_home(shift, ctrl)

        elif (key == Qt.Key_End) or ((key == Qt.Key_Down) and ctrl):
            self._key_end(shift, ctrl)

        elif key == Qt.Key_Up:
            if not self.is_cursor_on_last_line():
                self.set_cursor_position("eof")
            y_cursor = self.get_coordinates(cursor_position)[1]
            y_prompt = self.get_coordinates(self.current_prompt_pos)[1]
            if y_cursor > y_prompt:
                self.stdkey_up(shift)
            else:
                self.browse_history(backward=True)

        elif key == Qt.Key_Down:
            if not self.is_cursor_on_last_line():
                self.set_cursor_position("eof")
            y_cursor = self.get_coordinates(cursor_position)[1]
            y_end = self.get_coordinates("eol")[1]
            if y_cursor < y_end:
                self.stdkey_down(shift)
            else:
                self.browse_history(backward=False)

        elif key in (Qt.Key_PageUp, Qt.Key_PageDown):
            # XXX: Find a way to do this programmatically instead of calling
            # widget keyhandler (this won't work if the *event* is coming from
            # the event queue - i.e. if the busy buffer is ever implemented)
            ConsoleBaseWidget.keyPressEvent(self, event)

        elif key == Qt.Key_Escape and shift:
            self.clear_line()

        elif key == Qt.Key_Escape:
            self._key_escape()

        elif key == Qt.Key_L and ctrl:
            self.clear_terminal()

        elif key == Qt.Key_V and ctrl:
            self.paste()

        elif key == Qt.Key_X and ctrl:
            self.cut()

        elif key == Qt.Key_Z and ctrl:
            self.undo()

        elif key == Qt.Key_Y and ctrl:
            self.redo()

        elif key == Qt.Key_A and ctrl:
            self.selectAll()

        elif key == Qt.Key_Question and not self.has_selected_text():
            self._key_question(text)

        elif key == Qt.Key_ParenLeft and not self.has_selected_text():
            self._key_parenleft(text)

        elif key == Qt.Key_Period and not self.has_selected_text():
            self._key_period(text)

        elif len(text) and not self.isReadOnly():
            self.hist_wholeline = False
            self.insert_text(text)
            self._key_other(text)

        else:
            # Let the parent widget handle the key press event
            ConsoleBaseWidget.keyPressEvent(self, event)

    # ------ Key handlers
    def _key_enter(self):
        command = self.input_buffer
        self.insert_text("\n", at_end=True)
        self.on_enter(command)
        self.flush()

    def _key_other(self, text):
        raise NotImplementedError

    def _key_backspace(self, cursor_position):
        raise NotImplementedError

    def _key_tab(self):
        raise NotImplementedError

    def _key_ctrl_space(self):
        raise NotImplementedError

    def _key_home(self, shift, ctrl):
        if self.is_cursor_on_last_line():
            self.stdkey_home(shift, ctrl, self.current_prompt_pos)

    def _key_end(self, shift, ctrl):
        if self.is_cursor_on_last_line():
            self.stdkey_end(shift, ctrl)

    def _key_pageup(self):
        raise NotImplementedError

    def _key_pagedown(self):
        raise NotImplementedError

    def _key_escape(self):
        raise NotImplementedError

    def _key_question(self, text):
        raise NotImplementedError

    def _key_parenleft(self, text):
        raise NotImplementedError

    def _key_period(self, text):
        raise NotImplementedError

    # ------ History Management
    def load_history(self):
        """Load history from a .py file in user home directory"""
        if osp.isfile(self.history_filename):
            rawhistory, _ = encoding.readlines(self.history_filename)
            rawhistory = [line.replace("\n", "") for line in rawhistory]
            if rawhistory[1] != self.INITHISTORY[1]:
                rawhistory[1] = self.INITHISTORY[1]
        else:
            rawhistory = self.INITHISTORY
        history = [line for line in rawhistory if line and not line.startswith("#")]

        # Truncating history to X entries:
        while len(history) >= CONF.get("historylog", "max_entries"):
            del history[0]
            while rawhistory[0].startswith("#"):
                del rawhistory[0]
            del rawhistory[0]

        # Saving truncated history:
        try:
            encoding.writelines(rawhistory, self.history_filename)
        except EnvironmentError:
            pass

        return history

    # ------ Simulation standards input/output
    def write_error(self, text):
        """Simulate stderr"""
        self.flush()
        self.write(text, flush=True, error=True)
        if DEBUG:
            STDERR.write(text)

    def write(self, text, flush=False, error=False, prompt=False):
        """Simulate stdout and stderr"""
        if prompt:
            self.flush()
        if not isinstance(text, str):
            # This test is useful to discriminate QStrings from decoded str
            text = str(text)
        self.__buffer.append(text)
        ts = time.time()
        if flush or prompt:
            self.flush(error=error, prompt=prompt)
        elif ts - self.__timestamp > 0.05:
            self.flush(error=error)
            self.__timestamp = ts
            # Timer to flush strings cached by last write() operation in series
            self.__flushtimer.start(50)

    def flush(self, error=False, prompt=False):
        """Flush buffer, write text to console"""
        # Fix for Spyder Issue #2452
        try:
            text = "".join(self.__buffer)
        except TypeError:
            text = b"".join(self.__buffer)
            try:
                text = text.decode(locale.getpreferredencoding())
            except UnicodeDecodeError:
                pass
        self.__buffer = []
        try:
            self.insert_text(text, at_end=True, error=error, prompt=prompt)
            QCoreApplication.processEvents()
            self.repaint()
            # Clear input buffer:
            self.new_input_line = True
        except RuntimeError:
            # RuntimeError: wrapped C/C++ object of type QTextCursor has been deleted
            #
            # (This happens when the shell is closed while a thread is writing.
            # For example, when closing host application with an active logging
            # connected to the shell)
            pass

    # ------ Text Insertion
    def insert_text(self, text, at_end=False, error=False, prompt=False):
        """
        Insert text at the current cursor position
        or at the end of the command line
        """
        if at_end:
            # Insert text at the end of the command line
            self.append_text_to_shell(text, error, prompt)
        else:
            # Insert text at current cursor position
            ConsoleBaseWidget.insert_text(self, text)

    # ------ Re-implemented Qt Methods
    def focusNextPrevChild(self, next):
        """
        Reimplemented to stop Tab moving to the next window
        """
        if next:
            return False
        return ConsoleBaseWidget.focusNextPrevChild(self, next)

    # ------ Drag and drop
    def dragEnterEvent(self, event):
        """Drag and Drop - Enter event"""
        event.setAccepted(event.mimeData().hasFormat("text/plain"))

    def dragMoveEvent(self, event):
        """Drag and Drop - Move event"""
        if event.mimeData().hasFormat("text/plain"):
            event.setDropAction(Qt.MoveAction)
            event.accept()
        else:
            event.ignore()

    def dropEvent(self, event):
        """Drag and Drop - Drop event"""
        if event.mimeData().hasFormat("text/plain"):
            text = str(event.mimeData().text())
            if self.new_input_line:
                self.on_new_line()
            self.insert_text(text, at_end=True)
            self.setFocus()
            event.setDropAction(Qt.MoveAction)
            event.accept()
        else:
            event.ignore()

    def drop_pathlist(self, pathlist):
        """Drop path list"""
        raise NotImplementedError


# Example how to debug complex interclass call chains:
#
# from spyder.utils.debug import log_methods_calls
# log_methods_calls('log.log', ShellBaseWidget)


class PythonShellWidget(TracebackLinksMixin, ShellBaseWidget, GetHelpMixin):
    """Python shell widget"""

    QT_CLASS = ShellBaseWidget
    INITHISTORY = [
        "# -*- coding: utf-8 -*-",
        "# *** Spyder Python Console History Log ***",
    ]
    SEPARATOR = "%s##---(%s)---" % (os.linesep * 2, time.ctime())
    go_to_error = Signal(str)

    def __init__(self, parent, profile=False, initial_message=None, read_only=False):
        ShellBaseWidget.__init__(self, parent, profile, initial_message, read_only)
        TracebackLinksMixin.__init__(self)
        GetHelpMixin.__init__(self)

    # ------ Context menu
    def setup_context_menu(self):
        """Reimplements ShellBaseWidget method"""
        ShellBaseWidget.setup_context_menu(self)
        self.copy_without_prompts_action = create_action(
            self,
            _("Copy without prompts"),
            icon=get_icon("copywop.png"),
            triggered=self.copy_without_prompts,
        )
        actions = [self.copy_without_prompts_action]
        if not self.isReadOnly():
            clear_line_action = create_action(
                self,
                _("Clear line"),
                icon=get_icon("editdelete.png"),
                tip=_("Clear line"),
                triggered=self.clear_line,
            )
            clear_action = create_action(
                self,
                _("Clear shell"),
                icon=get_icon("editclear.png"),
                tip=_("Clear shell contents ('cls' command)"),
                triggered=self.clear_terminal,
            )
            actions += [clear_line_action, clear_action]
        add_actions(self.menu, actions)

    def contextMenuEvent(self, event):
        """Reimplements ShellBaseWidget method"""
        state = self.has_selected_text()
        self.copy_without_prompts_action.setEnabled(state)
        ShellBaseWidget.contextMenuEvent(self, event)

    @Slot()
    def copy_without_prompts(self):
        """Copy text to clipboard without prompts"""
        text = self.get_selected_text()
        lines = text.split(os.linesep)
        for index, line in enumerate(lines):
            if line.startswith(">>> ") or line.startswith("... "):
                lines[index] = line[4:]
        text = os.linesep.join(lines)
        QApplication.clipboard().setText(text)

    # ------ Key handlers
    def postprocess_keyevent(self, event):
        """Process keypress event"""
        ShellBaseWidget.postprocess_keyevent(self, event)
        if QToolTip.isVisible():
            _event, _text, key, _ctrl, _shift = restore_keyevent(event)
            self.hide_tooltip_if_necessary(key)

    def _key_other(self, text):
        """1 character key"""
        if self.is_completion_widget_visible():
            self.completion_text += text

    def _key_backspace(self, cursor_position):
        """Action for Backspace key"""
        if self.has_selected_text():
            self.check_selection()
            self.remove_selected_text()
        elif self.current_prompt_pos == cursor_position:
            # Avoid deleting prompt
            return
        elif self.is_cursor_on_last_line():
            self.stdkey_backspace()
            if self.is_completion_widget_visible():
                # Removing only last character because if there was a selection
                # the completion widget would have been canceled
                self.completion_text = self.completion_text[:-1]

    def _key_tab(self):
        """Action for TAB key"""
        if self.is_cursor_on_last_line():
            empty_line = not self.get_current_line_to_cursor().strip()
            if empty_line:
                self.stdkey_tab()
            else:
                self.show_code_completion(automatic=False)

    def _key_ctrl_space(self):
        """Action for Ctrl+Space"""
        if not self.is_completion_widget_visible():
            self.show_code_completion(automatic=False)

    def _key_pageup(self):
        """Action for PageUp key"""
        pass

    def _key_pagedown(self):
        """Action for PageDown key"""
        pass

    def _key_escape(self):
        """Action for ESCAPE key"""
        if self.is_completion_widget_visible():
            self.hide_completion_widget()

    def _key_question(self, text):
        """Action for '?'"""
        if self.get_current_line_to_cursor():
            last_obj = self.get_last_obj()
            if last_obj and not last_obj.isdigit():
                self.show_object_info(last_obj)
        self.insert_text(text)
        # In case calltip and completion are shown at the same time:
        if self.is_completion_widget_visible():
            self.completion_text += "?"

    def _key_parenleft(self, text):
        """Action for '('"""
        self.hide_completion_widget()
        if self.get_current_line_to_cursor():
            last_obj = self.get_last_obj()
            if last_obj and not last_obj.isdigit():
                self.insert_text(text)
                self.show_object_info(last_obj, call=True)
                return
        self.insert_text(text)

    def _key_period(self, text):
        """Action for '.'"""
        self.insert_text(text)
        if self.codecompletion_auto:
            # Enable auto-completion only if last token isn't a float
            last_obj = self.get_last_obj()
            if last_obj and not last_obj.isdigit():
                self.show_code_completion(automatic=True)

    # ------ Paste
    def paste(self):
        """Reimplemented slot to handle multiline paste action"""
        text = str(QApplication.clipboard().text())
        if len(text.splitlines()) > 1:
            # Multiline paste
            if self.new_input_line:
                self.on_new_line()
            self.remove_selected_text()  # Remove selection, eventually
            end = self.get_current_line_from_cursor()
            lines = self.get_current_line_to_cursor() + text + end
            self.clear_line()
            self.execute_lines(lines)
            self.move_cursor(-len(end))
        else:
            # Standard paste
            ShellBaseWidget.paste(self)

    # ------ Code Completion / Calltips
    # Methods implemented in child class:
    # (e.g. InternalShell)
    def get_dir(self, objtxt):
        """Return dir(object)"""
        raise NotImplementedError

    def get_module_completion(self, objtxt):
        """Return module completion list associated to object name"""
        pass

    def get_globals_keys(self):
        """Return shell globals() keys"""
        raise NotImplementedError

    def get_cdlistdir(self):
        """Return shell current directory list dir"""
        raise NotImplementedError

    def iscallable(self, objtxt):
        """Is object callable?"""
        raise NotImplementedError

    def get_arglist(self, objtxt):
        """Get func/method argument list"""
        raise NotImplementedError

    def get__doc__(self, objtxt):
        """Get object __doc__"""
        raise NotImplementedError

    def get_doc(self, objtxt):
        """Get object documentation dictionary"""
        raise NotImplementedError

    def get_source(self, objtxt):
        """Get object source"""
        raise NotImplementedError

    def is_defined(self, objtxt, force_import=False):
        """Return True if object is defined"""
        raise NotImplementedError

    def show_code_completion(self, automatic):
        """Display a completion list based on the current line"""
        # Note: unicode conversion is needed only for ExternalShellBase
        text = str(self.get_current_line_to_cursor())
        last_obj = self.get_last_obj()

        if not text:
            return

        if text.startswith("import "):
            # pylint: disable=assignment-from-no-return
            obj_list = self.get_module_completion(text)
            words = text.split(" ")
            if "," in words[-1]:
                words = words[-1].split(",")
            self.show_completion_list(
                obj_list, completion_text=words[-1], automatic=automatic
            )
            return

        elif text.startswith("from "):
            # pylint: disable=assignment-from-no-return
            obj_list = self.get_module_completion(text)
            if obj_list is None:
                return
            words = text.split(" ")
            if "(" in words[-1]:
                words = words[:-2] + words[-1].split("(")
            if "," in words[-1]:
                words = words[:-2] + words[-1].split(",")
            self.show_completion_list(
                obj_list, completion_text=words[-1], automatic=automatic
            )
            return

        obj_dir = self.get_dir(last_obj)
        if last_obj and obj_dir and text.endswith("."):
            self.show_completion_list(obj_dir, automatic=automatic)
            return

        # Builtins and globals
        if (
            not text.endswith(".")
            and last_obj
            and re.match(r"[a-zA-Z_0-9]*$", last_obj)
        ):
            b_k_g = dir(builtins) + self.get_globals_keys() + keyword.kwlist
            for objname in b_k_g:
                if objname.startswith(last_obj) and objname != last_obj:
                    self.show_completion_list(
                        b_k_g, completion_text=last_obj, automatic=automatic
                    )
                    return
            else:
                return

        # Looking for an incomplete completion
        if last_obj is None:
            last_obj = text
        dot_pos = last_obj.rfind(".")
        if dot_pos != -1:
            if dot_pos == len(last_obj) - 1:
                completion_text = ""
            else:
                completion_text = last_obj[dot_pos + 1 :]
                last_obj = last_obj[:dot_pos]
            completions = self.get_dir(last_obj)
            if completions is not None:
                self.show_completion_list(
                    completions, completion_text=completion_text, automatic=automatic
                )
                return

        # Looking for ' or ": filename completion
        q_pos = max([text.rfind("'"), text.rfind('"')])
        if q_pos != -1:
            completions = self.get_cdlistdir()
            if completions:
                self.show_completion_list(
                    completions, completion_text=text[q_pos + 1 :], automatic=automatic
                )
            return

    # ------ Drag'n Drop
    def drop_pathlist(self, pathlist):
        """Drop path list"""
        if pathlist:
            files = ["r'%s'" % path for path in pathlist]
            if len(files) == 1:
                text = files[0]
            else:
                text = "[" + ", ".join(files) + "]"
            if self.new_input_line:
                self.on_new_line()
            self.insert_text(text)
            self.setFocus()



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""
Text encoding utilities, text file I/O

Functions 'get_coding', 'decode', 'encode'  come from Eric4
source code (Utilities/__init___.py) Copyright © 2003-2009 Detlev Offenbach
"""

from __future__ import annotations

import os
import re
from codecs import BOM_UTF8, BOM_UTF16, BOM_UTF32

# ------------------------------------------------------------------------------
#  Functions for encoding and decoding *text data* itself, usually originating
#  from or destined for the *contents* of a file.
# ------------------------------------------------------------------------------

# Codecs for working with files and text.
CODING_RE = re.compile(r"coding[:=]\s*([-\w_.]+)")
CODECS = [
    "utf-8",
    "iso8859-1",
    "iso8859-15",
    "ascii",
    "koi8-r",
    "cp1251",
    "koi8-u",
    "iso8859-2",
    "iso8859-3",
    "iso8859-4",
    "iso8859-5",
    "iso8859-6",
    "iso8859-7",
    "iso8859-8",
    "iso8859-9",
    "iso8859-10",
    "iso8859-13",
    "iso8859-14",
    "latin-1",
    "utf-16",
]


def get_coding(text: str) -> str | None:
    """
    Function to get the coding of a text.
    @param text text to inspect (string)
    @return coding string
    """
    for line in text.splitlines()[:2]:
        try:
            result = CODING_RE.search(str(line))
        except UnicodeDecodeError:
            # This could fail because str assume the text
            # is utf8-like and we don't know the encoding to give
            # it to str
            pass
        else:
            if result:
                codec = result.group(1)
                # sometimes we find a false encoding that can
                # result in errors
                if codec in CODECS:
                    return codec


def decode(text: bytes) -> tuple[str, str] | tuple[str, str] | tuple[str, str]:
    """
    Function to decode a text.
    @param text text to decode (bytes)
    @return decoded text and encoding
    """
    try:
        if text.startswith(BOM_UTF8):
            # UTF-8 with BOM
            return str(text[len(BOM_UTF8) :], "utf-8"), "utf-8-bom"
        elif text.startswith(BOM_UTF16):
            # UTF-16 with BOM
            return str(text[len(BOM_UTF16) :], "utf-16"), "utf-16"
        elif text.startswith(BOM_UTF32):
            # UTF-32 with BOM
            return str(text[len(BOM_UTF32) :], "utf-32"), "utf-32"
        coding = get_coding(text)
        if coding:
            return str(text, coding), coding
    except (UnicodeError, LookupError):
        pass
    # Assume UTF-8
    try:
        return str(text, "utf-8"), "utf-8-guessed"
    except (UnicodeError, LookupError):
        pass
    # Assume Latin-1 (behaviour before 3.7.1)
    return str(text, "latin-1"), "latin-1-guessed"


def encode(text: str, orig_coding: str) -> tuple[bytes, str] | tuple[bytes, str]:
    """
    Function to encode a text.
    @param text text to encode (string)
    @param orig_coding type of the original coding (string)
    @return encoded text and encoding
    """
    if orig_coding == "utf-8-bom":
        return BOM_UTF8 + text.encode("utf-8"), "utf-8-bom"

    # Try saving with original encoding
    if orig_coding:
        try:
            return text.encode(orig_coding), orig_coding
        except (UnicodeError, LookupError):
            pass

    # Try declared coding spec
    coding = get_coding(text)
    if coding:
        try:
            return text.encode(coding), coding
        except (UnicodeError, LookupError):
            raise RuntimeError("Incorrect encoding (%s)" % coding)
    if (
        orig_coding
        and orig_coding.endswith("-default")
        or orig_coding.endswith("-guessed")
    ):
        coding = orig_coding.replace("-default", "")
        coding = orig_coding.replace("-guessed", "")
        try:
            return text.encode(coding), coding
        except (UnicodeError, LookupError):
            pass

    # Try saving as ASCII
    try:
        return text.encode("ascii"), "ascii"
    except UnicodeError:
        pass

    # Save as UTF-8 without BOM
    return text.encode("utf-8"), "utf-8"


def write(text: str, filename: str, encoding: str = "utf-8", mode: str = "wb") -> str:
    """
    Write 'text' to file ('filename') assuming 'encoding'
    Return (eventually new) encoding
    """
    text, encoding = encode(text, encoding)
    with open(filename, mode) as textfile:
        textfile.write(text)
    return encoding


def writelines(
    lines: list[str], filename: str, encoding: str = "utf-8", mode: str = "wb"
) -> str:
    """
    Write 'lines' to file ('filename') assuming 'encoding'
    Return (eventually new) encoding
    """
    return write(os.linesep.join(lines), filename, encoding, mode)


def read(filename: str, encoding: str = "utf-8") -> tuple[str, str]:
    """
    Read text from file ('filename')
    Return text and encoding
    """
    with open(filename, "rb") as file:
        text, encoding = decode(file.read())
    return text, encoding


def readlines(filename: str, encoding: str = "utf-8") -> tuple[list[str], str]:
    """
    Read lines from file ('filename')
    Return lines and encoding
    """
    text, encoding = read(filename, encoding)
    return text.split(os.linesep), encoding



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License


"""
Tests for dataframeeditor.py
"""

# guitest: show

import numpy as np
import pytest

try:
    import pandas as pd
    import pandas.testing as pdt

    from guidata.widgets.dataframeeditor import DataFrameEditor
except ImportError:
    # pandas is not installed
    pd = pdt = DataFrameEditor = None

from guidata.env import execenv
from guidata.qthelpers import exec_dialog, qt_app_context


@pytest.mark.skipif(pd is None, reason="pandas is not installed")
def test_dataframeeditor():
    """DataFrame editor test"""

    def test_edit(data, title="", parent=None):
        """Test subroutine"""
        dlg = DataFrameEditor(parent=parent)

        if dlg.setup_and_check(data, title=title):
            exec_dialog(dlg)
            return dlg.get_value()
        else:
            import sys

            sys.exit(1)

    with qt_app_context():
        df1 = pd.DataFrame(
            [
                [True, "bool"],
                [1 + 1j, "complex"],
                ["test", "string"],
                [1.11, "float"],
                [1, "int"],
                [np.random.rand(3, 3), "Unkown type"],
                ["Large value", 100],
                ["áéí", "unicode"],
            ],
            index=["a", "b", np.nan, np.nan, np.nan, "c", "Test global max", "d"],
            columns=[np.nan, "Type"],
        )
        out = test_edit(df1)
        pdt.assert_frame_equal(df1, out)

        result = pd.Series([True, "bool"], index=[np.nan, "Type"], name="a")
        out = test_edit(df1.iloc[0])
        pdt.assert_series_equal(result, out)

        df1 = pd.DataFrame(np.random.rand(100100, 10))
        out = test_edit(df1)
        pdt.assert_frame_equal(out, df1)

        series = pd.Series(np.arange(10), name=0)
        out = test_edit(series)
        pdt.assert_series_equal(series, out)

        execenv.print("OK")


if __name__ == "__main__":
    test_dataframeeditor()



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause
# (see guidata/LICENSE for details)

"""
Generate a dataset class from a dictionary
"""

# guitest: show

import numpy as np

from guidata.dataset import create_dataset_from_dict
from guidata.env import execenv

TEST_DICT1 = {
    "a": 1,
    "b": 2.0,
    "c": "test",
    "d": {"x": 1, "y": 3},
    "data": np.array([1, 2, 3]),
}
TEST_DICT2 = {
    "a": 1,
    "unsupported": [2.0, 3.0],
}


def test_dataset_from_dict():
    """Test generate dataset class from a dictionary"""
    for dictionary in (TEST_DICT1,):
        execenv.print(dictionary)
        dataset = create_dataset_from_dict(dictionary)
        execenv.print(dataset)
        execenv.print(dataset.create())
        execenv.print("")
    try:
        create_dataset_from_dict(TEST_DICT2)
        assert False, "Should have raised a ValueError"
    except ValueError:
        # This is expected
        pass


if __name__ == "__main__":
    test_dataset_from_dict()



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause
# (see guidata/LICENSE for details)

"""
dockable
--------

The `dockable` module provides a mixin class for widgets that can be docked
into a QMainWindow.
"""

from __future__ import annotations

from qtpy.QtCore import Qt
from qtpy.QtWidgets import QDockWidget, QWidget


class DockableWidgetMixin:
    """Mixin class for widgets that can be docked into a QMainWindow"""

    ALLOWED_AREAS = Qt.AllDockWidgetAreas
    LOCATION = Qt.TopDockWidgetArea
    FEATURES = (
        QDockWidget.DockWidgetClosable
        | QDockWidget.DockWidgetFloatable
        | QDockWidget.DockWidgetMovable
    )

    def __init__(self):
        self._isvisible = False
        self.dockwidget: QDockWidget | None = None
        self._allowed_areas = self.ALLOWED_AREAS
        self._location = self.LOCATION
        self._features = self.FEATURES

    @property
    def parent_widget(self) -> QWidget | None:
        """Return associated QWidget parent"""
        return self.parent()

    def setup_dockwidget(
        self,
        location: Qt.DockWidgetArea | None = None,
        features: QDockWidget.DockWidgetFeatures | None = None,
        allowed_areas: Qt.DockWidgetAreas | None = None,
    ) -> None:
        """Setup dockwidget parameters

        Args:
            location (Qt.DockWidgetArea): Dockwidget location
            features (QDockWidget.DockWidgetFeatures): Dockwidget features
            allowed_areas (Qt.DockWidgetAreas): Dockwidget allowed areas
        """
        assert self.dockwidget is None, (
            "Dockwidget must be setup before calling 'create_dockwidget'"
        )
        if location is not None:
            self._location = location
        if features is not None:
            self._features = features
        if allowed_areas is not None:
            self._allowed_areas = allowed_areas

    def get_focus_widget(self) -> QWidget | None:
        """Return widget to focus when dockwidget is visible"""
        return None

    def create_dockwidget(self, title: str) -> tuple[QDockWidget, Qt.DockWidgetArea]:
        """Add to parent QMainWindow as a dock widget

        Args:
            title (str): Dockwidget title

        Returns:
            tuple[QDockWidget, Qt.DockWidgetArea]: Dockwidget and location
        """
        dock = QDockWidget(title, self.parent_widget)
        dock.setObjectName(self.__class__.__name__ + "_dw")
        dock.setAllowedAreas(self._allowed_areas)
        dock.setFeatures(self._features)
        dock.setWidget(self)
        dock.visibilityChanged.connect(self.visibility_changed)
        self.dockwidget = dock
        return (dock, self._location)

    def is_visible(self) -> bool:
        """Return dockwidget visibility state"""
        return self._isvisible

    def visibility_changed(self, enable: bool) -> None:
        """DockWidget visibility has changed

        Args:
            enable (bool): Dockwidget visibility state
        """
        if enable:
            self.dockwidget.raise_()
            widget = self.get_focus_widget()  # pylint: disable=assignment-from-none
            if widget is not None:
                widget.setFocus()
        self._isvisible = enable and self.dockwidget.isVisible()


class DockableWidget(QWidget, DockableWidgetMixin):
    """Dockable widget

    Args:
        parent (QWidget): Parent widget
    """

    def __init__(self, parent: QWidget):
        QWidget.__init__(self, parent)
        DockableWidgetMixin.__init__(self)



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause
# (see guidata/LICENSE for details)
#
# The array editor subpackage was derived from Spyder's arrayeditor.py module
# which is licensed under the terms of the MIT License (see spyder/__init__.py
# for details), copyright © Spyder Project Contributors

# Note: string and unicode data types will be formatted with '%s' (see below)

"""Basic utilitarian functions and variables for the various array editor classes"""

import numpy as np

SUPPORTED_FORMATS = {
    "single": "%.6g",
    "double": "%.6g",
    "float_": "%.6g",
    "longfloat": "%.6g",
    "float16": "%.6g",
    "float32": "%.6g",
    "float64": "%.6g",
    "float96": "%.6g",
    "float128": "%.6g",
    "csingle": "%r",
    "complex_": "%r",
    "clongfloat": "%r",
    "complex64": "%r",
    "complex128": "%r",
    "complex192": "%r",
    "complex256": "%r",
    "byte": "%d",
    "bytes8": "%s",
    "short": "%d",
    "intc": "%d",
    "int_": "%d",
    "longlong": "%d",
    "intp": "%d",
    "int8": "%d",
    "int16": "%d",
    "int32": "%d",
    "int64": "%d",
    "ubyte": "%d",
    "ushort": "%d",
    "uintc": "%d",
    "uint": "%d",
    "ulonglong": "%d",
    "uintp": "%d",
    "uint8": "%d",
    "uint16": "%d",
    "uint32": "%d",
    "uint64": "%d",
    "bool_": "%r",
    "bool8": "%r",
    "bool": "%r",
}


LARGE_SIZE = 5e5
LARGE_NROWS = 1e5
LARGE_COLS = 60


# ==============================================================================
# Utility functions
# ==============================================================================
def is_float(dtype: np.dtype) -> bool:
    """Return True if datatype dtype is a float kind

    Args:
        dtype: numpy datatype

    Returns:
        True if dtype is a float kind
    """
    return ("float" in dtype.name) or dtype.name in ["single", "double"]


def is_number(dtype: np.dtype) -> bool:
    """Return True is datatype dtype is a number kind

    Args:
        dtype: numpy datatype

    Returns:
        True if dtype is a number kind
    """
    return (
        is_float(dtype)
        or ("int" in dtype.name)
        or ("long" in dtype.name)
        or ("short" in dtype.name)
    )


def get_idx_rect(index_list: list) -> tuple:
    """Extract the boundaries from a list of indexes

    Args:
        index_list: list of indexes

    Returns:
        tuple: (min_row, max_row, min_col, max_col)
    """
    rows, cols = list(zip(*[(i.row(), i.column()) for i in index_list]))
    return (min(rows), max(rows), min(cols), max(cols))



# -*- coding: utf-8 -*-
#
# Copyright (c) 2025, Codra, Pierre Raybaut.
# Licensed under the terms of the BSD 3-Clause

"""
cleanup
=======

Module for cleaning up build artifacts, cache files, and other temporary files
from Python projects. This module provides functionality to clean various types
of files generated during development, testing, and building processes.

This module can be used programmatically or as a command-line tool.
"""

from __future__ import annotations

import argparse
import os
import re
import shutil
import sys
from pathlib import Path


def get_project_root(start_path: Path | str | None = None) -> Path:
    """Find the project root directory by looking for pyproject.toml.

    Args:
        start_path: Starting path to search from. If None, uses current directory.

    Returns:
        The project root directory containing pyproject.toml.

    Raises:
        FileNotFoundError: If pyproject.toml is not found in the directory tree.
    """
    if start_path is None:
        start_path = Path.cwd()
    elif isinstance(start_path, str):
        start_path = Path(start_path)

    current = start_path.resolve()

    # Look for pyproject.toml in current directory and parent directories
    while current != current.parent:
        if (current / "pyproject.toml").exists():
            return current
        current = current.parent

    # Check root directory
    if (current / "pyproject.toml").exists():
        return current

    raise FileNotFoundError(
        f"pyproject.toml not found in {start_path} or any parent directory"
    )


def get_lib_name(project_root: Path) -> str:
    """Extract library name from pyproject.toml.

    Args:
        project_root: The root directory of the project.

    Returns:
        The library name extracted from pyproject.toml.

    Raises:
        FileNotFoundError: If pyproject.toml is not found.
        ValueError: If the 'name' field cannot be found in pyproject.toml.
    """
    pyproject_path = project_root / "pyproject.toml"

    if not pyproject_path.exists():
        raise FileNotFoundError(f"pyproject.toml not found in {project_root}")

    with open(pyproject_path, "r", encoding="utf-8") as f:
        content = f.read()

    # Look for name = "..." in [project] section
    match = re.search(
        r'\[project\].*?name\s*=\s*["\']([^"\']+)["\']', content, re.DOTALL
    )
    if match:
        return match.group(1)

    raise ValueError("Could not find 'name' field in pyproject.toml [project] section")


def get_mod_name(project_root: Path) -> str:
    """Get the main module name by looking for directories in the project.

    Args:
        project_root: The root directory of the project.

    Returns:
        The main module name (typically the library name with hyphens replaced by '_').
    """
    lib_name = get_lib_name(project_root)
    # Convert hyphens to underscores for module name
    mod_name = lib_name.replace("-", "_")

    # Check if a directory with this name exists
    if (project_root / mod_name).exists():
        return mod_name

    # Otherwise return the lib_name as is
    return lib_name


def remove_if_exists(path: Path) -> None:
    """Remove a file or directory if it exists.

    Args:
        path: Path to the file or directory to remove.
    """
    if path.exists():
        if path.is_dir():
            print(f"    Removing directory: {path}")
            shutil.rmtree(path)
        else:
            print(f"    Removing file: {path}")
            path.unlink()


def remove_glob_pattern(pattern: str, search_root: Path) -> None:
    """Remove files matching a glob pattern.

    Args:
        pattern: Glob pattern to match files/directories.
        search_root: Root directory to search from.
    """
    matches = list(search_root.glob(pattern))

    if matches:
        print(f"    Removing {len(matches)} items matching pattern: {pattern}")
        for match in matches:
            try:
                if match.is_dir():
                    if match.name == "__pycache__":
                        print(f"      Removing __pycache__ directory: {match}")
                    shutil.rmtree(match)
                else:
                    match.unlink()
            except (OSError, PermissionError) as e:
                print(f"      ⚠️ Warning: Could not remove {match}: {e}")


def clean_python_cache(project_root: Path) -> None:
    """Remove Python cache files and directories.

    Args:
        project_root: The root directory of the project.
    """
    print("  Cleaning Python cache files...")

    # Remove .pyc and .pyo files
    remove_glob_pattern("**/*.pyc", project_root)
    remove_glob_pattern("**/*.pyo", project_root)

    # Remove __pycache__ directories
    for pycache_dir in project_root.rglob("__pycache__"):
        if pycache_dir.is_dir():
            remove_if_exists(pycache_dir)


def clean_build_artifacts(project_root: Path, lib_name: str, mod_name: str) -> None:
    """Remove build artifacts and egg-info directories.

    Args:
        project_root: The root directory of the project.
        lib_name: The library name from pyproject.toml.
        mod_name: The main module name.
    """
    print("  Cleaning build artifacts...")

    # Remove egg-info directories
    remove_if_exists(project_root / f"{lib_name}.egg-info")
    remove_if_exists(project_root / f"{mod_name}.egg-info")

    # Remove build/dist directories and MANIFEST
    remove_if_exists(project_root / "MANIFEST")
    remove_if_exists(project_root / "build")
    remove_if_exists(project_root / "dist")
    remove_if_exists(project_root / "doc" / "_build")


def clean_coverage_files(project_root: Path) -> None:
    """Remove coverage-related files.

    Args:
        project_root: The root directory of the project.
    """
    print("  Cleaning coverage files...")

    remove_if_exists(project_root / ".coverage")
    remove_if_exists(project_root / "coverage.xml")
    remove_if_exists(project_root / "htmlcov")
    remove_if_exists(project_root / "sitecustomize.py")
    remove_if_exists(project_root / ".pytest_cache")

    # Remove .coverage.* files
    remove_glob_pattern(".coverage.*", project_root)


def clean_profile_files(project_root: Path) -> None:
    """Remove profiling-related files.

    Args:
        project_root: The root directory of the project.
    """
    print("  Cleaning profile files...")

    remove_glob_pattern("*.prof", project_root)
    remove_glob_pattern("*.prof.*", project_root)
    prof_dir = project_root / "prof"
    if prof_dir.exists():
        remove_glob_pattern("*.prof", prof_dir)
        remove_glob_pattern("*.prof.*", prof_dir)
        remove_glob_pattern("*.svg", prof_dir)


def clean_backup_files(project_root: Path) -> None:
    """Remove backup files and version control leftovers.

    Args:
        project_root: The root directory of the project.
    """
    print("  Cleaning backup files and version control leftovers...")
    remove_glob_pattern("**/*.bak", project_root)
    remove_glob_pattern("**/*~", project_root)
    remove_glob_pattern("**/*.orig", project_root)


def clean_log_files(project_root: Path) -> None:
    """Remove log files.

    Args:
        project_root: The root directory of the project.
    """
    print("  Cleaning log files...")
    remove_glob_pattern("**/*.log", project_root)


def clean_localization_files(project_root: Path, mod_name: str) -> None:
    """Remove localization template files.

    Args:
        project_root: The root directory of the project.
        mod_name: The main module name.
    """
    print("  Cleaning localization files...")
    remove_if_exists(project_root / "doc" / "locale" / "pot")
    remove_glob_pattern(f"{mod_name}/locale/{mod_name}.pot", project_root)


def clean_documentation_files(project_root: Path, lib_name: str, mod_name: str) -> None:
    """Remove documentation generation artifacts.

    Args:
        project_root: The root directory of the project.
        lib_name: The library name from pyproject.toml.
        mod_name: The main module name.
    """
    print("  Cleaning documentation files...")
    remove_if_exists(project_root / "doc" / "changelog.md")
    remove_if_exists(project_root / "doc" / "auto_examples")
    remove_if_exists(project_root / "doc" / "sg_execution_times.rst")
    remove_glob_pattern(f"{mod_name}/data/doc/{lib_name}*.pdf", project_root)


def clean_wix_installer_files(project_root: Path, lib_name: str) -> None:
    """Remove WiX installer related files.

    Args:
        project_root: The root directory of the project.
        lib_name: The library name from pyproject.toml.
    """
    print("  Cleaning WiX installer files...")
    wix_dir = project_root / "wix"
    if wix_dir.exists():
        remove_if_exists(wix_dir / "bin")
        remove_if_exists(wix_dir / "obj")
        remove_glob_pattern("*.bmp", wix_dir)
        remove_glob_pattern("*.wixpdb", wix_dir)
        remove_glob_pattern(f"{lib_name}*.wxs", wix_dir)


def clean_empty_directories(project_root: Path) -> None:
    """Remove empty directories left by version control branch switching.

    Args:
        project_root: The root directory of the project.
    """
    print("  Cleaning empty directories...")

    # Multiple passes to handle nested empty directories
    total_removed = 0
    max_passes = 10  # Prevent infinite loops

    for pass_num in range(max_passes):
        removed_this_pass = 0

        # Get all directories, sorted by depth (deepest first)
        all_dirs = []
        for dir_path in project_root.rglob("*"):
            if (
                dir_path.is_dir()
                and dir_path != project_root
                and not str(dir_path).endswith(".git")
                and not str(dir_path).endswith(".venv")
                and "site-packages" not in str(dir_path)
            ):
                all_dirs.append(dir_path)

        # Sort by depth (deepest first) to remove nested empty dirs
        all_dirs.sort(key=lambda x: len(x.parts), reverse=True)

        for empty_dir in all_dirs:
            try:
                if empty_dir.exists() and not any(empty_dir.iterdir()):
                    empty_dir.rmdir()
                    removed_this_pass += 1
                    total_removed += 1
            except OSError:
                # Directory might not be empty or permission issue
                pass

        # If no directories were removed this pass, we're done
        if removed_this_pass == 0:
            break

    if total_removed > 0:
        print(f"    ✓ Removed {total_removed} empty directories")


def clean_public_repo_dirs(project_root: Path, lib_name: str) -> None:
    """Remove directories related to public repository upload.

    Args:
        project_root: The root directory of the project.
        lib_name: The library name from pyproject.toml.
    """
    print("  Cleaning public repository directories...")

    parent_dir = project_root.parent
    temp_dir = parent_dir / f"{lib_name}_temp"
    public_dir = parent_dir / f"{lib_name}_public"

    remove_if_exists(temp_dir)
    remove_if_exists(public_dir)


def run_cleanup(project_root: Path | str | None = None) -> None:
    """Run all cleanup operations on a project.

    Args:
        project_root: The root directory of the project. If None, will search
                     for the project root starting from the current directory.
    """
    try:
        # Get project information
        if project_root is None:
            project_root = get_project_root()
        elif isinstance(project_root, str):
            project_root = Path(project_root)

        print(f"🧹 Cleaning up repository: {project_root}")

        lib_name = get_lib_name(project_root)
        mod_name = get_mod_name(project_root)

        print(f"Library name: {lib_name}")
        print(f"Module name: {mod_name}")

        # Change to project root directory
        original_cwd = os.getcwd()
        try:
            os.chdir(project_root)

            # Perform cleanup operations
            clean_build_artifacts(project_root, lib_name, mod_name)
            clean_python_cache(project_root)
            clean_public_repo_dirs(project_root, lib_name)
            clean_coverage_files(project_root)
            clean_profile_files(project_root)
            clean_backup_files(project_root)
            clean_log_files(project_root)
            clean_localization_files(project_root, mod_name)
            clean_documentation_files(project_root, lib_name, mod_name)
            clean_wix_installer_files(project_root, lib_name)
            clean_empty_directories(project_root)

            print("🆗 Cleanup completed successfully!")

        finally:
            os.chdir(original_cwd)

    except Exception as e:
        print(f"Error during cleanup: {e}", file=sys.stderr)
        raise


def main() -> None:
    """Main function for command-line interface."""
    parser = argparse.ArgumentParser(
        description="Clean up build artifacts, cache files, "
        "and temporary files from Python projects."
    )
    parser.add_argument(
        "path",
        nargs="?",
        default=None,
        help="Path to the project root directory "
        "(default: search from current directory)",
    )
    parser.add_argument(
        "--verbose", "-v", action="store_true", help="Enable verbose output"
    )

    args = parser.parse_args()

    try:
        run_cleanup(args.path)
    except Exception as e:
        if args.verbose:
            raise
        else:
            print(f"Error: {e}", file=sys.stderr)
            sys.exit(1)


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License


"""
Tests for codeeditor.py
"""

# guitest: show

from guidata.configtools import get_icon
from guidata.env import execenv
from guidata.qthelpers import qt_app_context
from guidata.widgets import codeeditor


def test_codeeditor():
    """Test Code editor."""
    with qt_app_context(exec_loop=True):
        widget = codeeditor.CodeEditor(language="python")
        widget.set_text_from_file(codeeditor.__file__)
        widget.resize(800, 600)
        widget.setWindowTitle("Code editor")
        widget.setWindowIcon(get_icon("guidata.svg"))
        widget.show()
        execenv.print("OK")


if __name__ == "__main__":
    test_codeeditor()



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

# ruff: noqa

"""
guidata.widgets.codeeditor
==========================

This package provides an Editor widget based on QtGui.QPlainTextEdit.

.. autoclass:: CodeEditor
    :show-inheritance:
    :members:

"""

# %% This line is for cell execution testing
# pylint: disable=C0103
# pylint: disable=R0903
# pylint: disable=R0911
# pylint: disable=R0201

from __future__ import annotations

from qtpy.QtCore import QRect, QSize, Qt, QTimer, Signal
from qtpy.QtGui import QColor, QFont, QPainter
from qtpy.QtWidgets import QPlainTextEdit, QWidget

import guidata.widgets.syntaxhighlighters as sh
from guidata.config import CONF, _
from guidata.configtools import get_font, get_icon
from guidata.qthelpers import (
    add_actions,
    create_action,
    is_dark_theme,
    win32_fix_title_bar_background,
)
from guidata.utils import encoding
from guidata.widgets import about


class LineNumberArea(QWidget):
    """Line number area (on the left side of the text editor widget)

    Args:
        editor: CodeEditor widget
    """

    def __init__(self, editor: CodeEditor) -> None:
        QWidget.__init__(self, editor)
        self.code_editor = editor
        self.setMouseTracking(True)

    def sizeHint(self):
        """Override Qt method"""
        return QSize(self.code_editor.compute_linenumberarea_width(), 0)

    def paintEvent(self, event):
        """Override Qt method"""
        self.code_editor.linenumberarea_paint_event(event)

    def mouseMoveEvent(self, event):
        """Override Qt method"""
        self.code_editor.linenumberarea_mousemove_event(event)

    def mouseDoubleClickEvent(self, event):
        """Override Qt method"""
        self.code_editor.linenumberarea_mousedoubleclick_event(event)

    def mousePressEvent(self, event):
        """Override Qt method"""
        self.code_editor.linenumberarea_mousepress_event(event)

    def mouseReleaseEvent(self, event):
        """Override Qt method"""
        self.code_editor.linenumberarea_mouserelease_event(event)

    def wheelEvent(self, event):
        """Override Qt method"""
        self.code_editor.wheelEvent(event)


class CodeEditor(QPlainTextEdit):
    """Code editor widget

    Args:
        parent: Parent widget
        language: Language used for syntax highlighting
        font: Font used for the text
        columns: Number of columns
        rows: Number of rows
        inactivity_timeout: after this delay of inactivity (in milliseconds), the
         :py:attr:`CodeEditor.SIG_EDIT_STOPPED` signal is emitted
    """

    # To have these attrs when early viewportEvent's are triggered
    linenumberarea = None

    #: Signal emitted when text changes and the user stops typing for some time
    SIG_EDIT_STOPPED = Signal()

    LANGUAGES = {
        "python": sh.PythonSH,
        "cython": sh.CythonSH,
        "fortran77": sh.Fortran77SH,
        "fortran": sh.FortranSH,
        "idl": sh.IdlSH,
        "diff": sh.DiffSH,
        "gettext": sh.GetTextSH,
        "nsis": sh.NsisSH,
        "html": sh.HtmlSH,
        "yaml": sh.YamlSH,
        "cpp": sh.CppSH,
        "opencL": sh.OpenCLSH,
        "enaml": sh.EnamlSH,
        # Every other language
        None: sh.TextSH,
    }

    def __init__(
        self,
        parent: QWidget = None,
        language: str | None = None,
        font: QFont | None = None,
        columns: int | None = None,
        rows: int | None = None,
        inactivity_timeout: int = 1000,
    ) -> None:
        QPlainTextEdit.__init__(self, parent)

        self.visible_blocks = []
        self.highlighter = None
        self.normal_color = None
        self.sideareas_color = None
        self.linenumbers_color = None

        # Line number area management
        self.linenumbers_margin = True
        self.linenumberarea_enabled = None
        self.linenumberarea_pressed = None
        self.linenumberarea_released = None

        self.timer = QTimer()
        self.timer.setSingleShot(True)
        self.timer.setInterval(inactivity_timeout)
        # When the editor is destroyed, the timer is destroyed as well, so we do
        # not need to disconnect the timer from the SIG_EDIT_STOPPED signal.

        # But... we connect the timer to the SIG_EDIT_STOPPED signal directly:
        # we do not connect it to the `emit` method for two reasons.
        #
        # 1. The documented way to connect a signal to another signal is the
        #    following: `signal1.connect(signal2)`.
        #
        # 2. When the editor is destroyed, if the timer is connected to the `emit`
        #    method, the timer will try to call the `emit` method which is still bound
        #    to the destroyed editor. This will cause a crash, eventually (there is a
        #    time window between the destruction of the editor and the destruction of
        #    the timer, so the crash is not guaranteed to happen, but it is possible).
        #    On the other hand, if the timer is connected to the SIG_EDIT_STOPPED
        #    signal, when the timeout is reached, the connection will be handled by
        #    Qt and the `emit` method will not be called, so there will be no crash.
        self.timer.timeout.connect(self.SIG_EDIT_STOPPED)

        self.textChanged.connect(self.text_has_changed)

        self.setFocusPolicy(Qt.StrongFocus)
        self.setup(language=language, font=font, columns=columns, rows=rows)

        self.update_color_mode()

    def update_color_mode(self) -> None:
        """Update color mode according to the current theme"""
        win32_fix_title_bar_background(self)
        suffix = "dark" if is_dark_theme() else "light"
        color_scheme = CONF.get("color_schemes", "default/" + suffix)
        self.highlighter.set_color_scheme(color_scheme)
        self.highlighter.rehighlight()
        self.linenumbers_color = QColor(
            Qt.lightGray if is_dark_theme() else Qt.darkGray
        )
        self.normal_color = self.highlighter.get_foreground_color()
        self.sideareas_color = self.highlighter.get_sideareas_color()
        self.linenumberarea.update()

    def text_has_changed(self) -> None:
        """Text has changed: restart the timer to emit SIG_EDIT_STOPPED after a delay"""
        if self.timer.isActive():
            self.timer.stop()
        self.timer.start()

    def contextMenuEvent(self, event):
        """Override Qt method"""
        menu = self.createStandardContextMenu()
        about_action = create_action(
            self,
            _("About..."),
            icon=get_icon("guidata.svg"),
            triggered=about.show_about_dialog,
        )
        add_actions(menu, (None, about_action))
        menu.exec(event.globalPos())

    def setup(self, language=None, font=None, columns=None, rows=None):
        """Setup widget"""
        if font is None:
            font = get_font(CONF, "codeeditor")
        self.setFont(font)
        self.setup_linenumberarea()
        if language is not None:
            language = language.lower()
        highlighter_class = self.LANGUAGES.get(language)
        if highlighter_class is None:
            raise ValueError("Unsupported language '%s'" % language)
        self.highlighter = highlighter_class(self.document(), self.font())
        self.highlighter.rehighlight()
        self.normal_color = self.highlighter.get_foreground_color()
        self.sideareas_color = self.highlighter.get_sideareas_color()
        if columns is not None:
            self.set_minimum_width(columns)
        if rows is not None:
            self.set_minimum_height(rows)

    def set_minimum_width(self, columns):
        """Set widget minimum width to show the specified number of columns"""
        width = self.fontMetrics().width("9" * (columns + 8))
        self.setMinimumWidth(width)

    def set_minimum_height(self, rows):
        """Set widget minimum height to show the specified number of rows"""
        height = self.fontMetrics().height() * (rows + 1)
        self.setMinimumHeight(height)

    def setup_linenumberarea(self):
        """Setup widget"""
        self.linenumberarea = LineNumberArea(self)
        self.blockCountChanged.connect(self.update_linenumberarea_width)
        self.updateRequest.connect(self.update_linenumberarea)
        self.linenumberarea_pressed = -1
        self.linenumberarea_released = -1
        self.set_linenumberarea_enabled(True)
        self.update_linenumberarea_width()

    def set_text_from_file(self, filename):
        """Set the text of the editor from file *fname*"""
        text, _enc = encoding.read(filename)
        self.setPlainText(text)

    # -----linenumberarea
    def set_linenumberarea_enabled(self, state):
        """

        :param state:
        """
        self.linenumberarea_enabled = state
        self.linenumberarea.setVisible(state)
        self.update_linenumberarea_width()

    def get_linenumberarea_width(self):
        """Return current line number area width"""
        return self.linenumberarea.contentsRect().width()

    def compute_linenumberarea_width(self):
        """Compute and return line number area width"""
        if not self.linenumberarea_enabled:
            return 0
        digits = 1
        maxb = max(1, self.blockCount())
        while maxb >= 10:
            maxb /= 10
            digits += 1
        if self.linenumbers_margin:
            linenumbers_margin = 3 + self.fontMetrics().width("9" * digits)
        else:
            linenumbers_margin = 0
        return linenumbers_margin

    def update_linenumberarea_width(self, new_block_count=None):
        """
        Update line number area width.

        new_block_count is needed to handle blockCountChanged(int) signal
        """
        self.setViewportMargins(self.compute_linenumberarea_width(), 0, 0, 0)

    def update_linenumberarea(self, qrect, dy):
        """Update line number area"""
        if dy:
            self.linenumberarea.scroll(0, dy)
        else:
            self.linenumberarea.update(
                0, qrect.y(), self.linenumberarea.width(), qrect.height()
            )
        if qrect.contains(self.viewport().rect()):
            self.update_linenumberarea_width()

    def linenumberarea_paint_event(self, event):
        """Painting line number area"""
        painter = QPainter(self.linenumberarea)
        painter.fillRect(event.rect(), self.sideareas_color)
        # This is needed to make that the font size of line numbers
        # be the same as the text one when zooming
        # See Issues 2296 and 4811
        font = self.font()
        font_height = self.fontMetrics().height()

        active_block = self.textCursor().block()
        active_line_number = active_block.blockNumber() + 1

        for top, line_number, block in self.visible_blocks:
            if self.linenumbers_margin:
                if line_number == active_line_number:
                    font.setWeight(QFont.Bold)
                    painter.setFont(font)
                    painter.setPen(self.normal_color)
                else:
                    font.setWeight(QFont.Normal)
                    painter.setFont(font)
                    painter.setPen(self.linenumbers_color)

                painter.drawText(
                    0,
                    top,
                    self.linenumberarea.width(),
                    font_height,
                    Qt.AlignRight | Qt.AlignBottom,
                    str(line_number),
                )

    def __get_linenumber_from_mouse_event(self, event):
        """Return line number from mouse event"""
        block = self.firstVisibleBlock()
        line_number = block.blockNumber()
        top = self.blockBoundingGeometry(block).translated(self.contentOffset()).top()
        bottom = top + self.blockBoundingRect(block).height()

        while block.isValid() and top < event.pos().y():
            block = block.next()
            top = bottom
            bottom = top + self.blockBoundingRect(block).height()
            line_number += 1

        return line_number

    def linenumberarea_mousemove_event(self, event):
        """Handling line number area mouse move event"""
        line_number = self.__get_linenumber_from_mouse_event(event)
        block = self.document().findBlockByNumber(line_number - 1)
        data = block.userData()

        # this disables pyflakes messages if there is an active drag/selection
        # operation
        check = self.linenumberarea_released == -1
        if data and data.code_analysis and check:
            self.__show_code_analysis_results(line_number, data.code_analysis)

        if event.buttons() == Qt.LeftButton:
            self.linenumberarea_released = line_number
            self.linenumberarea_select_lines(
                self.linenumberarea_pressed, self.linenumberarea_released
            )

    def linenumberarea_mousedoubleclick_event(self, event):
        """Handling line number area mouse double-click event"""
        line_number = self.__get_linenumber_from_mouse_event(event)
        shift = event.modifiers() & Qt.ShiftModifier

    def linenumberarea_mousepress_event(self, event):
        """Handling line number area mouse double press event"""
        line_number = self.__get_linenumber_from_mouse_event(event)
        self.linenumberarea_pressed = line_number
        self.linenumberarea_released = line_number
        self.linenumberarea_select_lines(
            self.linenumberarea_pressed, self.linenumberarea_released
        )

    def linenumberarea_mouserelease_event(self, event):
        """Handling line number area mouse release event"""
        self.linenumberarea_released = -1
        self.linenumberarea_pressed = -1

    def linenumberarea_select_lines(self, linenumber_pressed, linenumber_released):
        """Select line(s) after a mouse press/mouse press drag event"""
        find_block_by_line_number = self.document().findBlockByLineNumber
        move_n_blocks = linenumber_released - linenumber_pressed
        start_line = linenumber_pressed
        start_block = find_block_by_line_number(start_line - 1)

        cursor = self.textCursor()
        cursor.setPosition(start_block.position())

        # Select/drag downwards
        if move_n_blocks > 0:
            for n in range(abs(move_n_blocks) + 1):
                cursor.movePosition(cursor.NextBlock, cursor.KeepAnchor)
        # Select/drag upwards or select single line
        else:
            cursor.movePosition(cursor.NextBlock)
            for n in range(abs(move_n_blocks) + 1):
                cursor.movePosition(cursor.PreviousBlock, cursor.KeepAnchor)

        # Account for last line case
        if linenumber_released == self.blockCount():
            cursor.movePosition(cursor.EndOfBlock, cursor.KeepAnchor)
        else:
            cursor.movePosition(cursor.StartOfBlock, cursor.KeepAnchor)

        self.setTextCursor(cursor)

    def resizeEvent(self, event):
        """Reimplemented Qt method to handle line number area resizing"""
        QPlainTextEdit.resizeEvent(self, event)
        cr = self.contentsRect()
        self.linenumberarea.setGeometry(
            QRect(cr.left(), cr.top(), self.compute_linenumberarea_width(), cr.height())
        )

    def paintEvent(self, event):
        """Overrides paint event to update the list of visible blocks"""
        self.update_visible_blocks(event)
        QPlainTextEdit.paintEvent(self, event)

    def update_visible_blocks(self, event):
        """Update the list of visible blocks/lines position"""
        self.visible_blocks[:] = []
        block = self.firstVisibleBlock()
        blockNumber = block.blockNumber()
        top = int(
            self.blockBoundingGeometry(block).translated(self.contentOffset()).top()
        )
        bottom = top + int(self.blockBoundingRect(block).height())
        ebottom_bottom = self.height()

        while block.isValid():
            visible = bottom <= ebottom_bottom
            if not visible:
                break
            if block.isVisible():
                self.visible_blocks.append((top, blockNumber + 1, block))
            block = block.next()
            top = bottom
            bottom = top + int(self.blockBoundingRect(block).height())
            blockNumber = block.blockNumber()


if __name__ == "__main__":
    from guidata.qthelpers import qt_app_context

    with qt_app_context(exec_loop=True):
        widget = CodeEditor(columns=80, rows=40)
        widget.set_text_from_file(__file__)
        widget.show()



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause
# (see guidata/LICENSE for details)

"""
DataSet class conversion/creation functions
===========================================

Update and restore datasets
---------------------------

.. autofunction:: guidata.dataset.update_dataset

.. autofunction:: guidata.dataset.restore_dataset

Create dataset classes
----------------------

.. autofunction:: guidata.dataset.create_dataset_from_func

.. autofunction:: guidata.dataset.create_dataset_from_dict
"""

from __future__ import annotations

import importlib
import inspect
from typing import TYPE_CHECKING, Any

import guidata.dataset.dataitems as gdi
import guidata.dataset.datatypes as gdt
from guidata.io.jsonfmt import JSONReader, JSONWriter

if TYPE_CHECKING:
    import guidata.dataset.datatypes as gdt

# ==============================================================================
# Updating, restoring datasets
# ==============================================================================


def update_dataset(
    dest: gdt.DataSet, source: Any | dict[str, Any], visible_only: bool = False
) -> None:
    """Update `dest` dataset items from `source` dataset.

    Args:
        dest (DataSet): The destination dataset object to update.
        source (Union[Any, Dict[str, Any]]): The source object or dictionary containing
           matching attribute names.
        visible_only (bool): If True, update only visible items. Defaults
           to False.

    For each DataSet item, the function will try to get the attribute
    of the same name from the source.

    If the attribute exists in the source object or the key exists in the dictionary,
    it will be set as the corresponding attribute in the destination dataset.

    Computed items are automatically skipped as they are read-only and their values
    are calculated automatically based on other dataset items.

    Returns:
        None
    """
    for item in dest._items:
        key = item._name
        if isinstance(item.get_prop("data", "computed", None), gdt.ComputedProp):
            continue  # Skip computed items
        if hasattr(source, key):
            try:
                hide = item.get_prop_value("display", source, "hide", False)
            except AttributeError:
                # FIXME: Remove this try...except
                hide = False
            if visible_only and hide:
                continue
            setattr(dest, key, getattr(source, key))
        elif isinstance(source, dict) and key in source:
            setattr(dest, key, source[key])


def restore_dataset(source: gdt.DataSet, dest: Any | dict[str, Any]) -> None:
    """Restore `dest` dataset items from `source` dataset.

    Args:
        source (DataSet): The source dataset object to restore from.
        dest (Union[Any, Dict[str, Any]]): The destination object or dictionary.

    This function is almost the same as `update_dataset` but requires
    the source to be a DataSet instead of the destination.

    Symmetrically from `update_dataset`, `dest` may also be a dictionary.

    Computed items are automatically skipped when restoring to another dataset object
    (since computed values should be recalculated), but are included when restoring
    to a dictionary (since dictionaries store all current values).

    Returns:
        None
    """
    for item in source._items:
        key = item._name
        value = getattr(source, key)
        if hasattr(dest, key):
            if isinstance(item.get_prop("data", "computed", None), gdt.ComputedProp):
                continue  # Skip computed items if destination is not a dictionary
            try:
                setattr(dest, key, value)
            except AttributeError:
                # This attribute is a property, skipping this iteration
                continue
        elif isinstance(dest, dict):
            dest[key] = value


# ==============================================================================
# Generating a dataset class from a function signature
# ==============================================================================


def get_arg_info(func) -> dict[str, tuple[Any, Any]]:
    """Returns a dictionary where keys are the function argument names
    and values are tuples containing (default argument value, argument data type).

    Note: If the argument has no default value, it will be set to None.
          If the argument has no data type annotation, it will be set to None.

    Args:
        func: The function to get argument info from.

    Returns:
        The argument info dictionary.
    """
    signature = inspect.signature(func)
    arg_info = {}
    for name, param in signature.parameters.items():
        default_value = param.default if param.default != param.empty else None
        data_type = param.annotation if param.annotation != param.empty else None
        arg_info[name] = (default_value, data_type)
    return arg_info


def __get_dataitem_from_type(data_type: Any) -> gdi.DataItem:
    """Returns a DataItem instance from a data type.

    Args:
        data_type: The data type to get the DataItem from.

    Returns:
        The DataItem.
    """
    if not isinstance(data_type, str):
        # In case we are not using "from __future__ import annotations"
        data_type = data_type.__name__
    data_type = data_type.split("[")[0].split(".")[-1]
    typemap = {
        "int": gdi.IntItem,
        "float": gdi.FloatItem,
        "bool": gdi.BoolItem,
        "str": gdi.StringItem,
        "dict": gdi.DictItem,
        "ndarray": gdi.FloatArrayItem,
    }
    ditem_klass = typemap.get(data_type)
    if ditem_klass is None:
        raise ValueError(f"Unsupported data type: {data_type}")
    return ditem_klass


def create_dataset_from_func(func) -> gdt.DataSet:
    """Creates a DataSet class from a function signature.

    Args:
        func: The function to create the DataSet from.

    Returns:
        The DataSet class.

    Note: Supported data types are: int, float, bool, str, dict, np.ndarray.
    """
    klassname = "".join([s.capitalize() for s in func.__name__.split("_")]) + "DataSet"
    arg_info = get_arg_info(func)
    klassattrs = {}
    for name, (default_value, data_type) in arg_info.items():
        if data_type is None:
            raise ValueError(f"Argument '{name}' has no data type annotation.")
        ditem = __get_dataitem_from_type(data_type)
        klassattrs[name] = ditem(name, default=default_value)
    return type(klassname, (gdt.DataSet,), klassattrs)


# ==============================================================================
# Generating a dataset class from a dictionary
# ==============================================================================


def create_dataset_from_dict(
    dictionary: dict[str, Any], klassname: str | None = None
) -> gdt.DataSet:
    """Creates a DataSet class from a dictionary.

    Args:
        dictionary: The dictionary to create the DataSet class from.
        klassname: The name of the DataSet class. If None, the name is 'DictDataSet'.

    Returns:
        The DataSet class.

    Note: Supported data types are: int, float, bool, str, dict, np.ndarray.
    """
    klassname = "DictDataSet" if klassname is None else klassname
    klassattrs = {}
    for name, value in dictionary.items():
        ditem = __get_dataitem_from_type(type(value))
        klassattrs[name] = ditem(name, default=value)
    return type(klassname, (gdt.DataSet,), klassattrs)


# ==============================================================================
# JSON serialization/deserialization of datasets
# ==============================================================================


def dataset_to_json(param: gdt.DataSet) -> str:
    """Serialize dataset to JSON string.

    Args:
        param: dataset (gdt.DataSet)

    Returns:
        JSON string representation of the dataset
    """
    writer = JSONWriter(None)  # No filename, we'll get JSON text
    # Store the class name so we can deserialize to the correct type
    writer.write(param.__class__.__module__, "class_module")
    writer.write(param.__class__.__name__, "class_name")
    param.serialize(writer)
    return writer.get_json()


def json_to_dataset(json_str: str) -> gdt.DataSet:
    """Deserialize dataset from JSON string.

    Args:
        json_str: JSON string representation

    Returns:
        Deserialized dataset object
    """
    reader = JSONReader(json_str)

    # Read the class information
    class_module = reader.read("class_module")
    class_name = reader.read("class_name")

    module = importlib.import_module(class_module)
    param_class = getattr(module, class_name)

    # Create instance and deserialize
    param: gdt.DataSet = param_class()
    param.deserialize(reader)
    return param



# -*- coding: utf-8 -*-
#
# Copyright (c) 2025, DataLab Platform Developers, BSD 3-Clause license,
# see LICENSE file.

"""
gitreport
=========

Module for collecting Git repository information from Python modules.
This utility is designed to be used by pytest configurations to display
Git status information for the main project and its dependencies during test runs.

The module provides a clean, unified way to:
- Detect which Python modules are installed from Git repositories
- Collect Git branch, commit hash, and commit message information
- Present this information in a consistent format for test reports

This is particularly useful in development environments and CI/CD pipelines
where dependencies might be installed from development branches of Git repositories.
"""

from __future__ import annotations

import os
import os.path as osp
import subprocess
from typing import Any


class GitRepositoryInfo:
    """Container for Git repository information."""

    def __init__(self, name: str, branch: str, commit: str, message: str) -> None:
        """Initialize Git repository information.

        Args:
            name: Human-readable name of the repository/module
            branch: Git branch name
            commit: Short commit hash
            message: First line of commit message (truncated if needed)
        """
        self.name = name
        self.branch = branch
        self.commit = commit
        self.message = message

    def __repr__(self) -> str:
        """Return string representation of the Git repository info."""
        return (
            f"GitRepositoryInfo(name={self.name!r}, "
            f"branch={self.branch!r}, commit={self.commit!r})"
        )


def get_git_info_for_modules(
    modules_config: list[tuple[str, Any, str | None]],
) -> dict[str, GitRepositoryInfo]:
    """Get Git information for specified modules that are Git repositories.

    Args:
        modules_config: List of tuples containing:
            - name: Human-readable name for the module
            - module: The imported Python module object
            - custom_path: Custom path to check (None for module introspection)

    Returns:
        Dictionary mapping module names to GitRepositoryInfo objects

    Example:
        >>> import mymodule
        >>> import othermodule
        >>> modules_config = [
        ...     ("MyProject", mymodule, "."),  # Current directory
        ...     ("OtherLib", othermodule, None),  # Use module.__file__
        ... ]
        >>> repos = get_git_info_for_modules(modules_config)
        >>> for name, info in repos.items():
        ...     print(f"{name}: {info.branch}@{info.commit}")

    This function:
    1. Iterates through the provided module configuration
    2. For each module, determines the repository path:
       - Uses custom_path if provided
       - Otherwise searches up from module.__file__ for .git directory
    3. Collects Git information (branch, commit, message) if .git exists
    4. Returns a clean dictionary of results
    """
    git_repos = {}
    original_cwd = os.getcwd()

    for name, module, custom_path in modules_config:
        try:
            repo_path = _find_repository_path(module, custom_path)
            if repo_path is None:
                continue

            # Get Git information
            git_info = _extract_git_information(repo_path)
            if git_info is not None:
                git_repos[name] = GitRepositoryInfo(name, *git_info)

        except (
            subprocess.CalledProcessError,
            FileNotFoundError,
            OSError,
            AttributeError,
        ):
            # Silently skip modules that can't be processed
            continue
        finally:
            os.chdir(original_cwd)

    return git_repos


def _find_repository_path(module: Any, custom_path: str | None) -> str | None:
    """Find the Git repository path for a module.

    Args:
        module: The Python module object
        custom_path: Custom path to use, or None for module introspection

    Returns:
        Path to the Git repository root, or None if not found
    """
    if custom_path:
        # Use custom path (typically for main project)
        repo_path = custom_path
    else:
        # Find module path and look for Git repo
        if not (hasattr(module, "__file__") and module.__file__):
            return None

        module_path = osp.dirname(module.__file__)
        repo_path = module_path

        # Walk up the directory tree looking for .git
        while repo_path and repo_path != osp.dirname(repo_path):
            if osp.exists(osp.join(repo_path, ".git")):
                break
            repo_path = osp.dirname(repo_path)
        else:
            return None  # No .git directory found

    # Final check that .git directory exists at the determined path
    if not osp.exists(osp.join(repo_path, ".git")):
        return None

    return repo_path


def _extract_git_information(repo_path: str) -> tuple[str, str, str] | None:
    """Extract Git branch, commit, and message from a repository.

    Args:
        repo_path: Path to the Git repository root

    Returns:
        Tuple of (branch, commit, message) or None if extraction fails
    """
    os.chdir(repo_path)

    # Get branch name
    branch = subprocess.check_output(
        ["git", "rev-parse", "--abbrev-ref", "HEAD"], text=True, encoding="utf-8"
    ).strip()

    # Get short commit hash
    commit = subprocess.check_output(
        ["git", "rev-parse", "--short", "HEAD"], text=True, encoding="utf-8"
    ).strip()

    # Get commit message (first line only, truncated if needed)
    message = subprocess.check_output(
        ["git", "log", "-1", "--pretty=%B"], text=True, encoding="utf-8"
    ).strip()

    if len(message.splitlines()) > 1:
        message = message.splitlines()[0]

    # Truncate long messages
    message = message[:60] + "[…]" if len(message) > 60 else message

    return branch, commit, message


def format_git_info_for_pytest(
    git_repos: dict[str, GitRepositoryInfo], main_project_name: str
) -> list[str]:
    """Format Git repository information for pytest report headers.

    Args:
        git_repos: Dictionary of Git repository information
        main_project_name: Name of the main project (to show first)

    Returns:
        List of formatted strings ready for pytest report header

    Example output:
        [
            "Git information:",
            "  MyProject - Branch: develop, Commit: abc1234",
            "    Message: Fix important bug",
            "  Dependencies:",
            "    guidata - Branch: main, Commit: def5678",
            "      Message: Update documentation"
        ]
    """
    if not git_repos:
        return []

    gitlist = ["Git information:"]

    # Show main project first if available
    if main_project_name in git_repos:
        info = git_repos[main_project_name]
        gitlist.append(f"  {info.name} - Branch: {info.branch}, Commit: {info.commit}")
        gitlist.append(f"    Message: {info.message}")

    # Show dependencies
    deps = [name for name in git_repos if name != main_project_name]
    if deps:
        if main_project_name in git_repos:
            gitlist.append("  Dependencies:")
            prefix = "    "
        else:
            prefix = "  "

        for name in sorted(deps):
            info = git_repos[name]
            gitlist.append(
                f"{prefix}{info.name} - Branch: {info.branch}, Commit: {info.commit}"
            )
            if info.message:
                gitlist.append(f"{prefix}  Message: {info.message}")

    return gitlist



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#    userconfig License Agreement (MIT License)
#    ------------------------------------------
#
#    Copyright © 2009-2012 Pierre Raybaut
#
#    Permission is hereby granted, free of charge, to any person
#    obtaining a copy of this software and associated documentation
#    files (the "Software"), to deal in the Software without
#    restriction, including without limitation the rights to use,
#    copy, modify, merge, publish, distribute, sublicense, and/or sell
#    copies of the Software, and to permit persons to whom the
#    Software is furnished to do so, subject to the following
#    conditions:
#
#    The above copyright notice and this permission notice shall be
#    included in all copies or substantial portions of the Software.
#
#    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#    EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
#    OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
#    NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
#    HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
#    WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
#    OTHER DEALINGS IN THE SOFTWARE.


"""
User configuration management
-----------------------------

The ``guidata.userconfig`` module provides user configuration file (.ini file)
management features based on ``ConfigParser`` (standard Python library).

It is the exact copy of the open-source package `userconfig` (MIT license).

This module provides the following functions and classes:

* :py:func:`get_home_dir`: return user home directory
* :py:func:`get_config_basedir`: return user configuration base directory
* :py:class:`UserConfig`: user configuration file management class

.. autofunction:: get_home_dir

.. autofunction:: get_config_basedir

.. autoclass:: UserConfig
    :members:
"""

import configparser as cp
import os
import os.path as osp
import re
import sys
import time
from typing import Any


def try_remove_file(path: str, wait_time: float = 0.05, retries: int = 20) -> None:
    """Try to remove a file, waiting if necessary.

    Args:
        path: The path of the file to remove.
        wait_time: The time in seconds to wait between checks.
        retries: The number of times to retry before giving up.

    Raises:
        IOError: If the file cannot be removed.
    """
    if os.path.isfile(path):
        attempt = 0
        while attempt < retries:
            try:
                os.remove(path)
                return
            except IOError:
                time.sleep(wait_time)
                attempt += 1
        raise IOError(f"Unable to remove file {path} due to file lock")


def get_home_dir() -> str:
    """Return user home directory"""
    try:
        # expanduser() returns a raw byte string which needs to be
        # decoded with the codec that the OS is using to represent
        # file paths.
        path = os.fsdecode(osp.expanduser("~"))
    except Exception:
        path = ""

    if osp.isdir(path):
        return path
    else:
        # Get home from alternative locations
        for env_var in ("HOME", "USERPROFILE", "TMP"):
            # os.environ.get() returns a raw byte string which needs to be
            # decoded with the codec that the OS is using to represent
            # environment variables.
            path = os.fsdecode(os.environ.get(env_var, ""))
            if osp.isdir(path):
                return path
            else:
                path = ""

        if not path:
            raise RuntimeError(
                "Please set the environment variable HOME to "
                "your user/home directory path."
            )


def get_config_basedir() -> str:
    """Return user configuration base directory."""
    if sys.platform.startswith("linux"):
        # Follow the XDG standard to save settings
        home = os.environ.get("XDG_CONFIG_HOME", "")
        if not home:
            home = osp.join(get_home_dir(), ".config")
        if not osp.isdir(home):
            os.makedirs(home)
        return home
    return get_home_dir()


class NoDefault:
    """
    Custom object used to explicitly mark that no default value were
    provided.
    """

    pass


class UserConfig(cp.ConfigParser):
    """
    UserConfig class, based on ConfigParser
    name: name of the config
    options: dictionary containing options *or* list of tuples
    (section_name, options)

    Note that "get" and "set" arguments number and type
    differ from the overriden methods
    """

    default_section_name = "main"

    def __init__(self, defaults):
        cp.ConfigParser.__init__(self)
        self.name = "none"
        self.raw = 0  # 0 = substitutions are enabled / 1 = raw config parser
        assert isinstance(defaults, dict)
        for _key, val in list(defaults.items()):
            assert isinstance(val, dict)
        if self.default_section_name not in defaults:
            defaults[self.default_section_name] = {}
        self.defaults = defaults
        self.reset_to_defaults(save=False)
        self.check_default_values()

    def update_defaults(self, defaults):
        """Update the default configuration

        :param defaults: dict section -> dict {option: default value}
        """
        for key, sectdict in list(defaults.items()):
            if key not in self.defaults:
                self.defaults[key] = sectdict
            else:
                self.defaults[key].update(sectdict)
        self.reset_to_defaults(save=False)

    def save(self):
        """Save the configuration."""
        # In any case, the resulting config is saved in config file:
        self.__save()

    def set_application(self, name, version, load=True, raw_mode=False):
        """
        Set the application name and version

        :param name: name of the application
        :param version: current version in format "X.Y.Z"
        :param load: If True, load the configuration from dict
        :param raw_mode: If True, enable raw mode of ConfigParser
        """
        self.name = name
        self.raw = 1 if raw_mode else 0
        if (version is not None) and (re.match(r"^(\d+).(\d+).(\d+)", version) is None):
            raise RuntimeError(
                f"Version number {version!r} is incorrect - must be in X.Y.Z format"
            )

        if load:
            # If config file already exists, it overrides Default options:
            self.__load()
            if version != self.get_version(version):
                # Version has changed -> overwriting .ini file
                self.reset_to_defaults(save=False)
                self.__remove_deprecated_options()
                # Set new version number
                self.set_version(version, save=False)
            if self.defaults is None:
                # If no defaults are defined, set .ini file settings as default
                self.set_as_defaults()

    def check_default_values(self):
        """Check the static options for forbidden data types"""
        errors = []

        def _check(key, value):
            if value is None:
                return
            if isinstance(value, dict):
                for k, v in list(value.items()):
                    _check(key + "{}", k)
                    _check(key + "/" + k, v)
            elif isinstance(value, (list, tuple)):
                for v in value:
                    _check(key + "[]", v)
            else:
                if not isinstance(value, (bool, int, float, str)):
                    errors.append(f"Invalid value for {key}: {value}")

        for name, section in list(self.defaults.items()):
            assert isinstance(name, str)
            for key, value in list(section.items()):
                _check(key, value)
        if errors:
            for err in errors:
                print(err)
            raise ValueError("Invalid default values")

    def get_version(self, version="0.0.0"):
        """Return configuration (not application!) version"""
        return self.get(self.default_section_name, "version", version)

    def set_version(self, version="0.0.0", save=True):
        """Set configuration (not application!) version"""
        self.set(self.default_section_name, "version", version, save=save)

    def __load(self):
        """
        Load config from the associated .ini file
        """
        try:
            self.read(self.filename(), encoding="utf-8")
        except cp.MissingSectionHeaderError:
            print("Warning: File contains no section headers.")

    def __remove_deprecated_options(self):
        """
        Remove options which are present in the .ini file but not in defaults
        """
        for section in self.sections():
            for option, _ in self.items(section, raw=self.raw):
                if self.get_default(section, option) is NoDefault:
                    self.remove_option(section, option)
                    if len(self.items(section, raw=self.raw)) == 0:
                        self.remove_section(section)

    def __save(self):
        """
        Save config into the associated .ini file
        """
        fname = self.filename()
        if osp.isfile(fname):
            try_remove_file(fname, wait_time=0.05, retries=20)
        os.makedirs(osp.dirname(fname), mode=0o700, exist_ok=True)
        with open(fname, "w", encoding="utf-8") as configfile:
            self.write(configfile)

    def get_path(self, basename: str) -> str:
        """Return filename path inside configuration directory"""
        config_dir = osp.join(get_config_basedir(), f".{self.name}")
        if not osp.isdir(config_dir):
            os.makedirs(config_dir)
        return osp.join(config_dir, basename)

    def filename(self) -> str:
        """Return configuration file name"""
        return self.get_path(f"{self.name}.ini")

    def cleanup(self):
        """
        Remove .ini file associated to config
        """
        os.remove(self.filename())

    def set_as_defaults(self):
        """
        Set defaults from the current config
        """
        self.defaults = {}
        for section in self.sections():
            secdict = {}
            for option, value in self.items(section, raw=self.raw):
                secdict[option] = value
            self.defaults[section] = secdict

    def reset_to_defaults(self, save=True, verbose=False):
        """
        Reset config to Default values
        """
        for section, options in list(self.defaults.items()):
            for option in options:
                value = options[option]
                self.__set(section, option, value, verbose)
        if save:
            self.__save()

    def __check_section_option(self, section, option):
        """
        Private method to check section and option types
        """
        if section is None:
            section = self.default_section_name
        elif not isinstance(section, str):
            raise RuntimeError("Argument 'section' must be a string")
        if not isinstance(option, str):
            raise RuntimeError("Argument 'option' must be a string")
        return section

    def get_default(self, section, option):
        """
        Get Default value for a given (section, option)

        Useful for type checking in 'get' method
        """
        section = self.__check_section_option(section, option)
        options = self.defaults.get(section, {})
        return options.get(option, NoDefault)

    def get(self, section, option, default: Any = NoDefault, raw=None, **kwargs):
        """
        Get an option
        section=None: attribute a default section name
        default: default value (if not specified, an exception
        will be raised if option doesn't exist)
        """
        if raw is None:
            raw = self.raw

        section = self.__check_section_option(section, option)

        if not self.has_section(section):
            if default is NoDefault:
                raise RuntimeError(f"Unknown section {section!r}")
            else:
                self.add_section(section)

        if not self.has_option(section, option):
            if default is NoDefault:
                raise RuntimeError(f"Unknown option {section!r}/{option!r}")
            else:
                self.set(section, option, default)
                return default

        value = cp.ConfigParser.get(self, section, option, raw=raw)

        default_value = self.get_default(section, option)
        if isinstance(default_value, bool):
            value = eval(value)
        elif isinstance(default_value, float):
            value = float(value)
        elif isinstance(default_value, int):
            value = int(value)
        elif isinstance(default_value, str):
            pass
        else:
            try:
                # lists, tuples, ...
                value = eval(value)
            except Exception:
                pass

        return value

    def get_section(self, section):
        """Returns configuration values of the given section.

        The returned dict includes unset default values.

        :param section: section name
        :return: dict option name -> value
        """
        sect = self.defaults.get(section, {}).copy()
        for opt in self.options(section):
            sect[opt] = self.get(section, opt)
        return sect

    def __set(self, section, option, value, verbose):
        """
        Private set method
        """
        if not self.has_section(section):
            self.add_section(section)
        if not isinstance(value, str):
            value = repr(value)
        if verbose:
            print("{}[ {} ] = {}".format(section, option, value))
        cp.ConfigParser.set(self, section, option, value)

    def set_default(self, section, option, default_value):
        """
        Set Default value for a given (section, option)
        -> called when a new (section, option) is set and no default exists
        """
        section = self.__check_section_option(section, option)
        options = self.defaults.setdefault(section, {})
        options[option] = default_value

    def set(self, section, option, value, verbose=False, save=True):
        """
        Set an option
        section=None: attribute a default section name
        """
        section = self.__check_section_option(section, option)
        default_value = self.get_default(section, option)
        if default_value is NoDefault:
            default_value = value
            self.set_default(section, option, default_value)
        if isinstance(default_value, bool):
            value = bool(value)
        elif isinstance(default_value, float):
            value = float(value)
        elif isinstance(default_value, int):
            value = int(value)
        elif not isinstance(default_value, str):
            value = repr(value)
        self.__set(section, option, value, verbose)
        if save:
            self.__save()

    def remove_section(self, section):
        """Remove the given section and save the configuration."""
        cp.ConfigParser.remove_section(self, section)
        self.__save()

    def remove_option(self, section, option):
        """
        Remove the given option from the given section
        and save the configuration.
        """
        cp.ConfigParser.remove_option(self, section, option)
        self.__save()



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause
# (see guidata/LICENSE for details)

"""
Update/Restore dataset from/to another dataset or dictionary
"""

# guitest: show

import numpy as np

from guidata.dataset.conv import restore_dataset, update_dataset
from guidata.tests.dataset.test_all_items import Parameters


def test_update_restore_dataset():
    """Test update/restore dataset from/to another dataset or dictionary"""
    dataset = Parameters()
    dsdict = {
        "integer": 1,
        "float_slider": 1.0,
        "bool1": True,
        "string": "test",
        "floatarray": np.array([1, 2, 3]),
        "dictionary": {"a": 1, "b": 2},
    }
    # Update dataset from dictionary
    update_dataset(dataset, dsdict)
    # Check dataset values
    assert dataset.integer == dsdict["integer"]
    assert dataset.float_slider == dsdict["float_slider"]
    assert dataset.bool1 == dsdict["bool1"]
    assert dataset.string == dsdict["string"]
    assert np.all(dataset.floatarray == dsdict["floatarray"])
    assert dataset.dictionary == dsdict["dictionary"]
    # Update dataset from another dataset
    dataset2 = Parameters()
    update_dataset(dataset2, dataset)
    # Check dataset values
    assert dataset2.integer == dataset.integer
    assert dataset2.float_slider == dataset.float_slider
    assert dataset2.bool1 == dataset.bool1
    assert dataset2.string == dataset.string
    assert np.all(dataset2.floatarray == dataset.floatarray)
    assert dataset2.dictionary == dataset.dictionary
    # Restore dataset from dictionary
    restore_dataset(dataset, dsdict)
    # Check dataset values
    assert dataset.integer == dsdict["integer"]
    assert dataset.float_slider == dsdict["float_slider"]
    assert dataset.bool1 == dsdict["bool1"]
    assert dataset.string == dsdict["string"]
    assert np.all(dataset.floatarray == dsdict["floatarray"])
    assert dataset.dictionary == dsdict["dictionary"]


if __name__ == "__main__":
    test_update_restore_dataset()



# -*- coding: utf-8 -*-

"""
guidata.widgets.console
=======================

This package provides a Python console widget.

.. autoclass:: Console
    :show-inheritance:
    :members:

.. autoclass:: DockableConsole
    :show-inheritance:
    :members:

"""

from qtpy.QtCore import Qt

from guidata.config import CONF
from guidata.configtools import get_font
from guidata.qthelpers import win32_fix_title_bar_background
from guidata.widgets.console.internalshell import InternalShell
from guidata.widgets.dockable import DockableWidgetMixin


class Console(InternalShell):
    """
    Python console that run an interactive shell linked to
    the running process.

    :param parent: parent Qt widget
    :param namespace: available python namespace when the console start
    :type namespace: dict
    :param message: banner displayed before the first prompt
    :param commands: commands run when the interpreter starts
    :param type commands: list of string
    :param multithreaded: multithreaded support
    """

    def __init__(
        self,
        parent=None,
        namespace=None,
        message=None,
        commands=None,
        multithreaded=True,
        debug=False,
    ):
        InternalShell.__init__(
            self,
            parent=parent,
            namespace=namespace,
            message=message,
            commands=commands or [],
            multithreaded=multithreaded,
            debug=debug,
        )
        win32_fix_title_bar_background(self)
        self.setup()

    def setup(self):
        """Setup the calltip widget and show the console once all
        internal handler are ready."""
        font = get_font(CONF, "console")
        font.setPointSize(10)
        self.set_font(font)
        self.set_codecompletion_auto(True)
        self.set_calltips(True)
        self.setup_completion(size=(300, 180), font=font)
        try:
            self.exception_occurred.connect(self.show_console)
        except AttributeError:
            pass

    def closeEvent(self, event):
        """Reimplement Qt base method"""
        InternalShell.closeEvent(self, event)
        self.exit_interpreter()
        event.accept()


class DockableConsole(Console, DockableWidgetMixin):
    """
    Dockable Python console that run an interactive shell linked to
    the running process.

    :param parent: parent Qt widget
    :param namespace: available python namespace when the console start
    :type namespace: dict
    :param message: banner displayed before the first prompt
    :param commands: commands run when the interpreter starts
    :param type commands: list of string
    """

    LOCATION = Qt.BottomDockWidgetArea

    def __init__(
        self, parent, namespace, message, commands=None, multithreaded=True, debug=False
    ):
        DockableWidgetMixin.__init__(self)
        Console.__init__(
            self,
            parent=parent,
            namespace=namespace,
            message=message,
            commands=commands or [],
            multithreaded=multithreaded,
            debug=debug,
        )

    def show_console(self):
        """Show the console widget."""
        self.dockwidget.raise_()
        self.dockwidget.show()



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause
# (see guidata/LICENSE for details)
#
# The array editor subpackage was derived from Spyder's arrayeditor.py module
# which is licensed under the terms of the MIT License (see spyder/__init__.py
# for details), copyright © Spyder Project Contributors

# pylint: disable=C0103
# pylint: disable=R0903
# pylint: disable=R0911
# pylint: disable=R0201

"""
Module that provides array editor dialog boxes to edit various types of NumPy arrays
"""

from __future__ import annotations

import sys
from typing import Generic

import numpy as np
from qtpy.QtCore import QModelIndex, Qt, Slot
from qtpy.QtWidgets import (
    QComboBox,
    QDialog,
    QGridLayout,
    QHBoxLayout,
    QLabel,
    QMessageBox,
    QPushButton,
    QSpinBox,
    QStackedWidget,
    QWidget,
)

from guidata.config import _
from guidata.configtools import get_icon
from guidata.qthelpers import win32_fix_title_bar_background
from guidata.widgets.arrayeditor import utils
from guidata.widgets.arrayeditor.arrayhandler import (
    AnySupportedArray,
    BaseArrayHandler,
    MaskedArrayHandler,
    RecordArrayHandler,
)
from guidata.widgets.arrayeditor.editorwidget import (
    BaseArrayEditorWidget,
    DataArrayEditorWidget,
    MaskArrayEditorWidget,
    MaskedArrayEditorWidget,
    RecordArrayEditorWidget,
)


class ArrayEditor(QDialog, Generic[AnySupportedArray]):
    """Array Editor Dialog

    Args:
        parent: Parent widget (default: None)
    """

    __slots__ = (
        "data",
        "is_record_array",
        "is_masked_array",
        "arraywidget",
        "arraywidgets",
        "stack",
        "layout",
        "btn_save_and_close",
        "btn_close",
        "dim_indexes",
        "last_dim",
    )
    _data: BaseArrayHandler | MaskedArrayHandler | RecordArrayHandler
    arraywidget: (
        BaseArrayEditorWidget
        | MaskArrayEditorWidget
        | DataArrayEditorWidget
        | RecordArrayEditorWidget
    )
    layout: QGridLayout

    def __init__(self, parent: QWidget = None) -> None:
        QDialog.__init__(self, parent)
        win32_fix_title_bar_background(self)

        # Destroying the C++ object right after closing the dialog box,
        # otherwise it may be garbage-collected in another QThread
        # (e.g. the editor's analysis thread in Spyder), thus leading to
        # a segmentation fault on UNIX or an application crash on Windows
        self.setAttribute(Qt.WidgetAttribute.WA_DeleteOnClose)

        self.is_record_array = False
        self.is_masked_array = False
        self.arraywidgets: list[BaseArrayEditorWidget] = []
        self.btn_save_and_close = None
        self.btn_close = None
        # Values for 3d array editor
        self.dim_indexes = [{}, {}, {}]
        self.last_dim = 0  # Adjust this for changing the startup dimension
        self.stack: QStackedWidget | None = None

    def setup_and_check(
        self,
        data: AnySupportedArray,
        title="",
        readonly=False,
        xlabels=None,
        ylabels=None,
        variable_size=False,
        add_title_suffix=True,
    ) -> bool:
        """
        Setup the array editor dialog box and check if the array is supported.

        Args:
            data: Array to edit
            title: Dialog box title
            readonly: Flag indicating if the array is read only
            xlabels: List of x labels
            ylabels: List of y labels
            variable_size: Flag indicating if the array is variable size
            add_title_suffix: Flag indicating if the array type should be added to the
             title

        Returns:
            True if the array is supported, False otherwise
        """
        readonly = readonly or not data.flags.writeable
        self._variable_size = (
            variable_size and not readonly and xlabels is None and ylabels is None
        )

        if readonly and variable_size:
            QMessageBox.warning(
                self,
                _("Conflicing edition flags"),
                _(
                    "Array editor was initialized in both readonly and variable "
                    "size mode."
                )
                + "\n"
                + _("The array editor will remain in readonly mode."),
            )

        if variable_size and (xlabels is not None or ylabels is not None):
            QMessageBox.warning(
                self,
                _("Unsupported array format"),
                _(
                    "Array editor does not support array with x/y labels in "
                    "variable size mode."
                )
                + "\n"
                + _("You will not be able to add or remove rows/columns."),
            )

        self.is_record_array = data.dtype.names is not None
        self.is_masked_array = isinstance(data, np.ma.MaskedArray)

        if self.is_masked_array:
            self._data = MaskedArrayHandler(data, self._variable_size)  # type: ignore
        elif self.is_record_array:
            self._data = RecordArrayHandler(data, self._variable_size)
        else:
            self._data = BaseArrayHandler(data, self._variable_size)

        if data.ndim > 3:
            self.error(_("Arrays with more than 3 dimensions are not supported"))
            return False
        if xlabels is not None and len(xlabels) != self._data.shape[1]:
            self.error(
                _("The 'xlabels' argument length do no match array column number")
            )
            return False
        if ylabels is not None and len(ylabels) != self._data.shape[0]:
            self.error(_("The 'ylabels' argument length do no match array row number"))
            return False
        if not self.is_record_array:
            dtn = data.dtype.name
            if (
                dtn not in utils.SUPPORTED_FORMATS
                and not dtn.startswith("str")
                and not dtn.startswith("unicode")
            ):
                arr_ = _("%s arrays") % data.dtype.name
                self.error(_("%s are currently not supported") % arr_)
                return False

        self.layout = QGridLayout()
        self.setLayout(self.layout)
        self.setWindowIcon(get_icon("arredit.png"))
        if add_title_suffix:
            title = (
                str(title) + " - " + _("NumPy array") if title else _("Array editor")
            )
            if readonly:
                title += " (" + _("read only") + ")"
        self.setWindowTitle(title)
        self.resize(600, 500)

        # Stack widget
        self.stack = QStackedWidget(self)
        if self.is_record_array:
            for name in data.dtype.names:
                w = RecordArrayEditorWidget(
                    self,
                    self._data,  # type: ignore
                    name,
                    readonly,
                    xlabels,
                    ylabels,
                    variable_size,
                )
                self.arraywidgets.append(w)
                self.stack.addWidget(w)
        elif self.is_masked_array:
            w1 = MaskedArrayEditorWidget(
                self, self._data, readonly, xlabels, ylabels, variable_size
            )
            self.arraywidgets.append(w1)
            self.stack.addWidget(w1)

            w2 = DataArrayEditorWidget(
                self,
                self._data,  # type: ignore
                readonly,
                xlabels,
                ylabels,
                variable_size,
            )
            self.arraywidgets.append(w2)
            self.stack.addWidget(w2)

            w3 = MaskArrayEditorWidget(
                self,
                self._data,
                readonly,
                xlabels,
                ylabels,
                variable_size,
            )
            self.arraywidgets.append(w3)
            self.stack.addWidget(w3)
        elif data.ndim == 3:
            pass
        else:
            w = BaseArrayEditorWidget(
                self, self._data, readonly, xlabels, ylabels, variable_size
            )
            self.stack.addWidget(w)
        self.arraywidget = self.stack.currentWidget()
        if self.arraywidget:
            self.arraywidget.model.dataChanged.connect(self.save_and_close_enable)
        for wdg in self.arraywidgets:
            wdg.model.SIZE_CHANGED.connect(self.update_all_tables_on_size_change)
        self.stack.currentChanged.connect(self.current_widget_changed)
        self.layout.addWidget(self.stack, 1, 0)

        # Buttons configuration
        btn_layout = QHBoxLayout()
        if self.is_record_array or self.is_masked_array or data.ndim == 3:
            if self.is_record_array:
                btn_layout.addWidget(QLabel(_("Record array fields:")))
                names = []
                for name in data.dtype.names:
                    field = data.dtype.fields[name]
                    text = name
                    if len(field) >= 3:
                        title = field[2]
                        if not isinstance(title, str):
                            title = repr(title)
                        text += " - " + title
                    names.append(text)
            else:
                names = [_("Masked data"), _("Data"), _("Mask")]
            if data.ndim == 3:
                # QSpinBox
                self.index_spin = QSpinBox(self, keyboardTracking=False)
                self.index_spin.valueChanged.connect(self.change_active_widget)
                # QComboBox
                names = [str(i) for i in range(3)]
                ra_combo = QComboBox(self)
                ra_combo.addItems(names)
                ra_combo.currentIndexChanged.connect(self.current_dim_changed)
                # Adding the widgets to layout
                label = QLabel(_("Axis:"))
                btn_layout.addWidget(label)
                btn_layout.addWidget(ra_combo)
                self.shape_label = QLabel()
                btn_layout.addWidget(self.shape_label)
                label = QLabel(_("Index:"))
                btn_layout.addWidget(label)
                btn_layout.addWidget(self.index_spin)
                self.slicing_label = QLabel()
                btn_layout.addWidget(self.slicing_label)
                # set the widget to display when launched
                self.current_dim_changed(self.last_dim)
            else:
                ra_combo = QComboBox(self)
                ra_combo.currentIndexChanged.connect(self.stack.setCurrentIndex)
                ra_combo.addItems(names)
                btn_layout.addWidget(ra_combo)
            if self.is_masked_array:
                label = QLabel(_("<u>Warning</u>: changes are applied separately"))
                label.setToolTip(
                    _(
                        "For performance reasons, changes applied "
                        "to masked array won't be reflected in "
                        "array's data (and vice-versa)."
                    )
                )
                btn_layout.addWidget(label)

        btn_layout.addStretch()

        if not readonly:
            self.btn_save_and_close = QPushButton(_("Save and Close"))
            self.btn_save_and_close.setDisabled(True)
            self.btn_save_and_close.clicked.connect(self.accept)
            btn_layout.addWidget(self.btn_save_and_close)

        self.btn_close = QPushButton(_("Close"))
        self.btn_close.setAutoDefault(True)
        self.btn_close.setDefault(True)
        self.btn_close.clicked.connect(self.reject)
        btn_layout.addWidget(self.btn_close)
        self.layout.addLayout(btn_layout, 2, 0)

        self.setMinimumSize(400, 300)

        # Make the dialog act as a window
        self.setWindowFlags(Qt.WindowType.Window)

        return True

    @Slot(bool, bool)
    def update_all_tables_on_size_change(self, rows: bool, cols: bool) -> None:
        """Updates all array editor widgets when rows and/or columns count changes.

        Args:
            rows: Flag to indicate the number of rows changed
            cols: Flag to indicate the number of columns changed
        """
        for wdg in self.arraywidgets:
            wdg.model.fetch(rows, cols)
            wdg.model.set_hue_values()
        if self._data.ndim == 3:
            self.current_dim_changed(self.last_dim)

    @Slot(QModelIndex, QModelIndex)
    def save_and_close_enable(
        self, _left_top: QModelIndex, _bottom_right: QModelIndex
    ) -> None:
        """Handle the data change event to enable the save and close button."""
        if self.btn_save_and_close:
            self.btn_save_and_close.setEnabled(True)
            self.btn_save_and_close.setAutoDefault(True)
            self.btn_save_and_close.setDefault(True)

    def current_widget_changed(self, index: int) -> None:
        """Handle the current widget change event to connect the dataChanged signal

        Args:
            index: Index of the current widget
        """
        if self.stack is not None:
            self.arraywidget = self.stack.widget(index)
            self.arraywidget.model.dataChanged.connect(self.save_and_close_enable)

    def change_active_widget(self, index: int) -> None:
        """This is implemented for handling negative values in index for
        3d arrays, to give the same behavior as slicing

        Args:
            index: Index of the current widget
        """
        string_index = [":"] * 3
        string_index[self.last_dim] = "<font color=red>%i</font>"
        self.slicing_label.setText(
            (r"Slicing: [" + ", ".join(string_index) + "]") % index
        )
        data_index = self._data.shape[self.last_dim] + index if index < 0 else index
        slice_index = [slice(None)] * 3
        slice_index[self.last_dim] = data_index

        stack_index = self.dim_indexes[self.last_dim].get(data_index)
        if stack_index is None and self.stack is not None:
            stack_index = self.stack.count()
            try:
                w = BaseArrayEditorWidget(
                    self,
                    self._data,
                    variable_size=self._variable_size,
                    current_slice=slice_index,
                )
                self.stack.addWidget(w)
            except IndexError:  # Handle arrays of size 0 in one axis
                w = BaseArrayEditorWidget(
                    self,
                    self._data,
                    variable_size=self._variable_size,
                    current_slice=slice_index,
                )
                self.stack.addWidget(w)
            self.arraywidgets.append(
                w
            )  # required to fetch the new columns/rows if added/deleted
            w.model.SIZE_CHANGED.connect(self.update_all_tables_on_size_change)
            self.dim_indexes[self.last_dim][data_index] = stack_index
            self.stack.update()
            self.stack.setCurrentIndex(stack_index)

    def current_dim_changed(self, index: int) -> None:
        """This change the active axis the array editor is plotting over
        in 3D

        Args:
            index: Index of the current widget
        """
        self.last_dim = index
        string_size = ["%i"] * 3
        string_size[index] = "<font color=red>%i</font>"
        self.shape_label.setText(
            ("Shape: (" + ", ".join(string_size) + ")    ") % self._data.shape
        )
        if self.index_spin.value() != 0:
            self.index_spin.setValue(0)
        else:
            # this is done since if the value is currently 0 it does not emit
            # currentIndexChanged(int)
            self.change_active_widget(0)
        self.index_spin.setRange(-self._data.shape[index], self._data.shape[index] - 1)

    @Slot()
    def accept(self) -> None:
        """Reimplement Qt method"""
        self._data.apply_changes()
        QDialog.accept(self)

    def get_value(self) -> AnySupportedArray:
        """Return modified array -- the returned array is a copy if variable size is
        True and readonly is False

        Returns:
            Modified array
        """
        # It is import to avoid accessing Qt C++ object as it has probably
        # already been destroyed, due to the Qt.WA_DeleteOnClose attribute
        self._data.reset_shape_if_changed()
        return self._data.get_array()

    def error(self, message: str) -> None:
        """An error occured, closing the dialog box

        Args:
            message: Error message
        """
        if "pytest" in sys.modules:
            raise RuntimeError(message)
        QMessageBox.critical(self, _("Array editor"), message)
        self.setAttribute(Qt.WidgetAttribute.WA_DeleteOnClose)
        self.reject()

    @Slot()
    def reject(self) -> None:
        """Reimplement Qt method"""
        self._data.clear_changes()
        QDialog.reject(self)


class BaseArrayEditor(ArrayEditor[np.ndarray]):
    """Optional wrapper class to get type inferance for normal numpy arrays"""


class RecordArrayEditor(ArrayEditor[np.ndarray]):
    """Optional wrapper class to get type inferance for record numpy arrays"""


class MaskedArrayEditor(ArrayEditor[np.ma.MaskedArray]):
    """Optional wrapper class to get type inferance for masked numpy arrays"""



# -----------------------------------------------------------------------------
#  Copyright (C) 2019 Alberto Sottile
#
#  Distributed under the terms of the 3-clause BSD License.
# -----------------------------------------------------------------------------

__version__ = "0.5.2"  # Patched (see pull request below)

import sys
import platform

if sys.platform == "darwin":
    from distutils.version import LooseVersion as V

    if V(platform.mac_ver()[0]) < V("10.14"):
        from ._dummy import *
    else:
        from ._mac_detect import *
    del V
# Patch: https://github.com/albertosottile/darkdetect/pull/21
elif (
    sys.platform == "win32"
    and platform.release().isdigit()
    and int(platform.release()) >= 10
):
    # Checks if running Windows 10 version 10.0.14393 (Anniversary Update) OR HIGHER. The getwindowsversion method returns a tuple.
    # The third item is the build number that we can use to check if the user has a new enough version of Windows.
    winver = int(platform.version().split(".")[2])
    if winver >= 14393:
        from ._windows_detect import *
    else:
        from ._dummy import *
elif sys.platform == "linux":
    from ._linux_detect import *
else:
    from ._dummy import *

del sys, platform



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause
# (see guidata/LICENSE for details)

"""
Data serialization and deserialization
--------------------------------------

The ``guidata.io`` package provides the core features for data
(:py:class:`guidata.dataset.DataSet` or other objects) serialization and
deserialization.

Base classes for I/O handlers
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

Base classes for writing custom readers and writers:

* :py:class:`BaseIOHandler`
* :py:class:`WriterMixin`

.. autoclass:: GroupContext

.. autoclass:: BaseIOHandler
    :members:

.. autoclass:: WriterMixin
    :members:

Configuration files (.ini)
^^^^^^^^^^^^^^^^^^^^^^^^^^

Reader and writer for the serialization of data sets into .ini files:

* :py:class:`INIReader`
* :py:class:`INIWriter`

.. autoclass:: INIReader
    :members:

.. autoclass:: INIWriter
    :members:

JSON files (.json)
^^^^^^^^^^^^^^^^^^

Reader and writer for the serialization of data sets into .json files:

* :py:class:`JSONReader`
* :py:class:`JSONWriter`

.. autoclass:: JSONReader
    :members:

.. autoclass:: JSONWriter
    :members:

HDF5 files (.h5)
^^^^^^^^^^^^^^^^

Reader and writer for the serialization of data sets into .h5 files:

* :py:class:`HDF5Reader`
* :py:class:`HDF5Writer`

.. autoclass:: HDF5Reader
    :members:

.. autoclass:: HDF5Writer
    :members:
"""

# pylint: disable=unused-import
from .base import BaseIOHandler, GroupContext, WriterMixin  # noqa
from .h5fmt import HDF5Handler, HDF5Reader, HDF5Writer  # noqa
from .inifmt import INIHandler, INIReader, INIWriter  # noqa
from .jsonfmt import JSONHandler, JSONReader, JSONWriter  # noqa



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause
# (see guidata/LICENSE for details)

"""
Base classes for I/O handlers
"""

from __future__ import annotations

import datetime
from collections.abc import Callable
from typing import Any

import numpy as np


class GroupContext:
    """
    Group context manager object.

    This class provides a context manager for managing a group within a handler.

    Args:
        handler (BaseIOHandler): The handler object. It should be an instance
            of a subclass of BaseIOHandler.
        group_name (str): The group name. Represents the name of the group in
            the context.
    """

    def __init__(self, handler: BaseIOHandler, group_name: str) -> None:
        """
        Initialization method for GroupContext.
        """
        self.handler = handler
        self.group_name = group_name

    def __enter__(self) -> GroupContext:
        """
        Enter the context.

        This method is called when entering the 'with' statement.
        It begins a new group on the handler.

        Returns:
            self (GroupContext): An instance of the GroupContext class.
        """
        self.handler.begin(self.group_name)
        return self

    def __exit__(
        self, exc_type: type | None, exc_value: Exception | None, traceback: Any | None
    ) -> bool:
        """
        Exit the context.

        This method is called when exiting the 'with' statement. It ends the group
        on the handler if no exception occurred within the context.

        Args:
            exc_type (Optional[type]): The type of exception that occurred, if any.
            exc_value (Optional[Exception]): The instance of Exception that occurred,
             if any.
            traceback (Optional[Any]): A traceback object encapsulating the call stack
             at the point where the exception originally occurred, if any.

        Returns:
            False (bool): A boolean False value indicating that the exception
             was not handled here.
        """
        if exc_type is None:
            self.handler.end(self.group_name)
        return False


class BaseIOHandler:
    """
    Base I/O Handler with group context manager.
    This class serves as the base class for I/O handlers. It provides methods for
    managing sections of a file, referred to as "groups", as well as context
    management for these groups.
    """

    def __init__(self) -> None:
        """
        Initialization method for BaseIOHandler.
        This method initializes the option list, which will be used to manage
        the current section or "group".
        """
        self.option = []

    def group(self, group_name: str) -> GroupContext:
        """
        Enter a group and return a context manager to be used with the `with` statement.
        Args:
            group_name (str): The name of the group to enter.
        Returns:
            GroupContext: A context manager for the group.
        """
        return GroupContext(self, group_name)

    def begin(self, section: str) -> None:
        """
        Begin a new section.
        This method is called when a new section is started. It adds the section to
        the list of options, which effectively makes it the current section.
        Args:
            section (str): The name of the section to begin.
        """
        self.option.append(section)

    def end(self, section: str) -> None:
        """
        End the current section.
        This method is called when a section is ended.
        It removes the section from the list of options,
        asserting it's the expected one, and moves to the previous section if any.
        Args:
            section (str): The name of the section to end.
        """
        sect = self.option.pop(-1)
        assert sect == section, (
            "Ending section does not match the current section: %s != %s"
            % (
                sect,
                section,
            )
        )


class WriterMixin:
    """
    Mixin class providing the write() method.
    This mixin class is intended to be used with classes that need to write
    different types of values.
    """

    def write(self, val: Any, group_name: str | None = None) -> None:
        """
        Write a value depending on its type, optionally within a named group.
        Args:
            val (Any): The value to be written.
            group_name (Optional[str]): The name of the group. If provided, the group
            context will be used for writing the value.
        """
        if group_name:
            self.begin(group_name)

        if isinstance(val, bool):
            self.write_bool(val)
        elif isinstance(val, int):
            self.write_int(val)
        elif isinstance(val, float):
            self.write_float(val)
        elif isinstance(val, str):
            self.write_any(val)
        elif isinstance(val, np.ndarray):
            self.write_array(val)
        elif np.isscalar(val):
            self.write_any(val)
        elif val is None:
            self.write_none()
        elif isinstance(val, (list, tuple)):
            self.write_sequence(val)
        elif isinstance(val, datetime.datetime):
            self.write_float(val.timestamp())
        elif isinstance(val, datetime.date):
            self.write_int(val.toordinal())
        elif hasattr(val, "serialize") and isinstance(val.serialize, Callable):
            # The object has a DataSet-like `serialize` method
            val.serialize(self)
        else:
            raise NotImplementedError(
                "cannot serialize %r of type %r" % (val, type(val))
            )

        if group_name:
            self.end(group_name)



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause
# (see guidata/LICENSE for details)

"""
Configuration related functions
-------------------------------

Access configured options
^^^^^^^^^^^^^^^^^^^^^^^^^

.. autofunction:: get_icon

.. autofunction:: get_image_file_path

.. autofunction:: get_image_label

.. autofunction:: get_image_layout

.. autofunction:: get_family

.. autofunction:: get_font

.. autofunction:: get_pen

.. autofunction:: get_brush

Add image paths
^^^^^^^^^^^^^^^

.. autofunction:: add_image_path

.. autofunction:: add_image_module_path
"""

from __future__ import annotations

import gettext
import os
import os.path as osp
import sys
import warnings
from collections.abc import Callable
from typing import TYPE_CHECKING

from guidata.utils.misc import decode_fs_string, get_module_path, get_system_lang

if TYPE_CHECKING:
    from qtpy import QtCore as QC
    from qtpy import QtGui as QG
    from qtpy import QtWidgets as QW

    from guidata.userconfig import UserConfig

IMG_PATH = []


def get_module_data_path(modname: str, relpath: str | None = None) -> str:
    """Return module *modname* data path
    Handles py2exe/cx_Freeze distributions

    Args:
        modname (str): module name
        relpath (str): relative path to module data directory

    Returns:
        str: module data path
    """
    datapath = getattr(sys.modules[modname], "DATAPATH", "")
    if not datapath:
        datapath = get_module_path(modname)
        parentdir = osp.normpath(osp.join(datapath, osp.pardir))
        if osp.isfile(parentdir):
            # Parent directory is not a directory but the 'library.zip' file:
            # this is either a py2exe or a cx_Freeze distribution
            datapath = osp.abspath(osp.join(osp.join(parentdir, osp.pardir), modname))
    if relpath is not None:
        datapath = osp.abspath(osp.join(datapath, relpath))
    return datapath


def get_translation(modname: str, dirname: str | None = None) -> Callable[[str], str]:
    """Return translation callback for module *modname*

    Args:
        modname (str): module name
        dirname (str): module directory

    Returns:
        Callable[[str], str]: translation callback
    """
    if dirname is None:
        dirname = modname
    # fixup environment var LANG in case it's unknown
    if "LANG" not in os.environ:
        lang = get_system_lang()
        if lang is not None:
            os.environ["LANG"] = lang
    modlocpath = get_module_locale_path(dirname)
    return gettext.translation(modname, modlocpath, fallback=True).gettext


def get_module_locale_path(modname: str) -> str:
    """Return module *modname* gettext translation path

    Args:
        modname (str): module name

    Returns:
        str: module gettext translation path
    """
    localepath = getattr(sys.modules[modname], "LOCALEPATH", "")
    if not localepath:
        localepath = get_module_data_path(modname, relpath="locale")
    return localepath


def add_image_path(path: str, subfolders: bool = True) -> None:
    """Append image path (opt. with its subfolders) to global list IMG_PATH

    Args:
        path (str): image path
        subfolders (bool): include subfolders
    """
    if not isinstance(path, str):
        path = decode_fs_string(path)
    global IMG_PATH
    IMG_PATH.append(path)
    if subfolders:
        for fileobj in os.listdir(path):
            pth = osp.join(path, fileobj)
            if osp.isdir(pth):
                IMG_PATH.append(pth)


def add_image_module_path(modname: str, relpath: str, subfolders: bool = True) -> None:
    """
    Appends image data path relative to a module name.
    Used to add module local data that resides in a module directory
    but will be shipped under sys.prefix / share/ ...

    modname must be the name of an already imported module as found in
    sys.modules

    Args:
        modname (str): module name
        relpath (str): relative path to module data directory
        subfolders (bool): include subfolders
    """
    add_image_path(get_module_data_path(modname, relpath=relpath), subfolders)


def get_image_file_path(name: str, default: str = "not_found.png") -> str:
    """
    Return the absolute path to image with specified name
    name, default: filenames with extensions

    Args:
        name (str): name of the image
        default (str): default image name. Defaults to "not_found.png".

    Raises:
        RuntimeError: if image file not found

    Returns:
        str: absolute path to image
    """
    for pth in IMG_PATH:
        full_path = osp.join(pth, name)
        if osp.isfile(full_path):
            return osp.abspath(full_path)
    if default is not None:
        try:
            return get_image_file_path(default, None)
        except RuntimeError:
            raise RuntimeError("Image file %r not found" % name)
    else:
        raise RuntimeError()


ICON_CACHE = {}


def get_std_icon(name: str) -> QG.QIcon | None:
    """Get standard icon by name"""
    # Importing Qt objects here because this module should not depend on them
    # pylint: disable=import-outside-toplevel

    # Try to get standard icon first
    from guidata.qthelpers import get_std_icon

    try:
        return get_std_icon(name)
    except AttributeError:
        return None


def get_icon(name: str, default: str = "not_found.png") -> QG.QIcon:
    """
    Construct a QIcon from the file with specified name
    name, default: filenames with extensions or standard Qt icon names

    Args:
        name (str): name of the icon
        default (str): default icon name. Defaults to "not_found.png".

    Returns:
        QG.QIcon: icon
    """
    try:
        return ICON_CACHE[name]
    except KeyError:
        std_icon = get_std_icon(name)
        if std_icon is not None:
            return std_icon

        std_default_icon = get_std_icon(default)

        # Importing Qt objects here because this module should not depend on them
        # pylint: disable=import-outside-toplevel

        # Retrieve icon from file (original implementation)
        from qtpy import QtGui as QG

        try:
            icon = QG.QIcon(get_image_file_path(name, default))
            ICON_CACHE[name] = icon
        except RuntimeError:  # default is a standard icon name
            if std_default_icon is not None:
                return std_default_icon
            raise
        return icon


def get_image_label(name, default="not_found.png") -> QW.QLabel:
    """
    Construct a QLabel from the file with specified name
    name, default: filenames with extensions

    Args:
        name (str): name of the icon
        default (str): default icon name. Defaults to "not_found.png".

    Returns:
        QW.QLabel: label
    """
    # Importing Qt here because this module should be independent from it
    from qtpy import QtGui as QG  # pylint: disable=import-outside-toplevel
    from qtpy import QtWidgets as QW  # pylint: disable=import-outside-toplevel

    label = QW.QLabel()
    pixmap = QG.QPixmap(get_image_file_path(name, default))
    label.setPixmap(pixmap)
    return label


def get_image_layout(
    imagename: str, text: str = "", tooltip: str = "", alignment: QC.Qt.Alignment = None
) -> tuple[QW.QHBoxLayout, QW.QLabel]:
    """
    Construct a QHBoxLayout including image from the file with specified name,
    left-aligned text [with specified tooltip]

    Args:
        imagename (str): name of the icon
        text (str): text to display. Defaults to "".
        tooltip (str): tooltip to display. Defaults to "".
        alignment (QC.Qt.Alignment): alignment of the text. Defaults to None.

    Returns:
        tuple[QW.QHBoxLayout, QW.QLabel]: layout, label
    """
    # Importing Qt here because this module should be independent from it
    from qtpy import QtCore as QC  # pylint: disable=import-outside-toplevel
    from qtpy import QtWidgets as QW  # pylint: disable=import-outside-toplevel

    if alignment is None:
        alignment = QC.Qt.AlignLeft
    layout = QW.QHBoxLayout()
    if alignment in (QC.Qt.AlignCenter, QC.Qt.AlignRight):
        layout.addStretch()
    layout.addWidget(get_image_label(imagename))
    label = QW.QLabel(text)
    label.setToolTip(tooltip)
    layout.addWidget(label)
    if alignment in (QC.Qt.AlignCenter, QC.Qt.AlignLeft):
        layout.addStretch()
    return (layout, label)


def font_is_installed(font: str) -> list[str]:
    """Check if font is installed

    Args:
        font (str): font name

    Returns:
        list[str]: list of installed fonts
    """
    # Importing Qt here because this module should be independent from it
    from qtpy import PYQT5
    from qtpy import QtGui as QG  # pylint: disable=import-outside-toplevel

    if PYQT5:
        fontfamilies = QG.QFontDatabase().families()
    else:
        # Qt6
        fontfamilies = QG.QFontDatabase.families()
    return [fam for fam in fontfamilies if str(fam) == font]


MONOSPACE = [
    "Cascadia Code PL",
    "Cascadia Mono PL",
    "Cascadia Code",
    "Cascadia Mono",
    "Consolas",
    "Lucida Console",
    "Lucida Sans Typewriter",
    "Monospace",
    "Menlo",
    "Courier New",
    "Courier",
    "Bitstream Vera Sans Mono",
    "Andale Mono",
    "Liberation Mono",
    "Monaco",
    "Fixedsys",
    "monospace",
    "Fixed",
    "Terminal",
]


def get_family(families: str | list[str]) -> str:
    """Return the first installed font family in family list

    Args:
        families (str|list[str]): font family or list of font families

    Returns:
        str: first installed font family
    """
    if not isinstance(families, list):
        families = [families]
    for family in families:
        if font_is_installed(family):
            return family
    else:
        # On Windows, in offscreen/headless mode, Qt may not detect fonts properly
        # Don't warn in this case, just return the first family as a fallback
        is_windows_offscreen = (
            sys.platform == "win32" and os.environ.get("QT_QPA_PLATFORM") == "offscreen"
        )
        if not is_windows_offscreen:
            warnings.warn("None of the following fonts is installed: %r" % families)
        return families[0] if families else ""


def get_font(conf: UserConfig, section: str, option: str = "") -> QG.QFont:
    """
    Construct a QFont from the specified configuration file entry
    conf: UserConfig instance
    section [, option]: configuration entry

    Args:
        conf (UserConfig): UserConfig instance
        section (str): configuration entry
        option (str): configuration entry. Defaults to "".

    Returns:
        QG.QFont: font
    """
    # Importing Qt here because this module should be independent from it
    from qtpy import QtGui as QG  # pylint: disable=import-outside-toplevel

    if not option:
        option = "font"
    if "font" not in option:
        option += "/font"
    font = QG.QFont()
    if conf.has_option(section, option + "/family/nt"):
        families = conf.get(section, option + "/family/" + os.name)
    elif conf.has_option(section, option + "/family"):
        families = conf.get(section, option + "/family")
    else:
        families = None
    if families is not None:
        if not isinstance(families, list):
            families = [families]
        family = None
        for family in families:
            if font_is_installed(family):
                break
        font.setFamily(family)
    if conf.has_option(section, option + "/size"):
        font.setPointSize(conf.get(section, option + "/size"))
    if conf.get(section, option + "/bold", False):
        font.setWeight(QG.QFont.Bold)
    else:
        font.setWeight(QG.QFont.Normal)
    return font


def get_pen(
    conf: UserConfig,
    section: str,
    option: str = "",
    color: str = "black",
    width: int = 1,
    style: str = "SolidLine",
) -> QG.QPen:
    """
    Construct a QPen from the specified configuration file entry
    conf: UserConfig instance
    section [, option]: configuration entry
    [color]: default color
    [width]: default width
    [style]: default style

    Args:
        conf (UserConfig): UserConfig instance
        section (str): configuration entry
        option (str): configuration entry. Defaults to "".
        color (str): default color. Defaults to "black".
        width (int): default width. Defaults to 1.
        style (str): default style. Defaults to "SolidLine".

    Returns:
        QG.QPen: pen
    """
    # Importing Qt here because this module should be independent from it
    from qtpy import QtCore as QC  # pylint: disable=import-outside-toplevel
    from qtpy import QtGui as QG  # pylint: disable=import-outside-toplevel

    if "pen" not in option:
        option += "/pen"
    color = conf.get(section, option + "/color", color)
    color = QG.QColor(color)
    width = conf.get(section, option + "/width", width)
    style_name = conf.get(section, option + "/style", style)
    style = getattr(QC.Qt, style_name)
    return QG.QPen(color, width, style)


def get_brush(
    conf: UserConfig,
    section: str,
    option: str = "",
    color: str = "black",
    alpha: float = 1.0,
) -> QG.QBrush:
    """
    Construct a QBrush from the specified configuration file entry
    conf: UserConfig instance
    section [, option]: configuration entry
    [color]: default color
    [alpha]: default alpha-channel

    Args:
        conf (UserConfig): UserConfig instance
        section (str): configuration entry
        option (str): configuration entry. Defaults to "".
        color (str): default color. Defaults to "black".
        alpha (float): default alpha-channel. Defaults to 1.0.

    Returns:
        QG.QBrush: brush
    """
    # Importing Qt here because this module should be independent from it
    from qtpy import QtGui as QG  # pylint: disable=import-outside-toplevel

    if "brush" not in option:
        option += "/brush"
    color = conf.get(section, option + "/color", color)
    color = QG.QColor(color)
    alpha = conf.get(section, option + "/alphaF", alpha)
    color.setAlphaF(alpha)
    return QG.QBrush(color)
    return QG.QBrush(color)
